import sys
from datetime import datetime
import logging
import os
import textwrap
import shutil
import pprint

import click
import matplotlib.pyplot as plt
import pandas as pd
import geopandas as gpd

import dew_gwdata as gd  # Access to SA Geodata database
import wrap_technote as tn  # GW data analysis functions


from pandas.plotting import register_matplotlib_converters

register_matplotlib_converters()


logger = tn.get_logger()


@click.command()
@click.option("-p", "--reporting-period", default=tn.CURRENT_RPERIOD)
@click.option("-v", "--verbose", count=True)
@click.option("-s1", "--step1", is_flag=True)
@click.option("-s2", "--step2", is_flag=True)
@click.option("-s3", "--step3", is_flag=True)
@click.option("-s", "--static", is_flag=True)
@click.option("-r", "--report", default="")
@click.option("--figures/--no-figures", is_flag=True, default=True)
@click.argument("resource", required=False)
def waterlevels(
    resource, reporting_period, verbose, step1, step2, step3, static, report, figures
):
    return waterlevels_run(
        resource,
        reporting_period,
        verbose,
        step1,
        step2,
        step3,
        static,
        report,
        figures,
    )


def waterlevels_run(
    resource,
    reporting_period,
    verbose=0,
    step1=True,
    step2=True,
    step3=True,
    static=True,
    report="",
    figures=True,
):
    """Run the water level data analysis for a WRAP resource.

    Args:
        resource (wrap_tn.Resource): the resource to analyse
        reporting_period (wrap_tn.ReportingPeriod): the reporting period
            to analyse the resource for
        verbose (int): pre-configure loguru logging messages, use
            1 for INFO to console, and 2 for INFO to console & DEBUG to
            a file in the current directory wraptn_waterlevels.log.
            Set to anything else to use whatever loguru is configured for
            via tn.get_logger().configure()
        step1 (bool): execute step 1 of the analysis (see
            :func:`wrap_technote.scripts.step1_data_validation` below)
        step2 (bool): execute step 1 of the analysis (see XXX below)
        step3 (bool): execute step 1 of the analysis (see XXX below)
        static (bool): run the code to create the static site resources
            (see XXX below)
        report (str): ?
        figures (bool): create figures, yes or no. You would only select
            False in order to speed up the execution time of this function.

    """
    handlers = []
    if verbose == 1:
        handlers.append({"sink": sys.stdout, "level": "INFO"})
    if verbose == 2:
        handlers.append({"sink": sys.stdout, "level": "INFO"})
        handlers.append({"sink": "wraptn_waterlevels.log", "level": "DEBUG"})
    config = {
        "handlers": handlers,
    }
    logger.configure(**config)

    logger.info(f"Using dew_gwdata {gd.__version__}")
    logger.info(f"Using wrap_technote {tn.__version__}")

    plt.rcParams["figure.dpi"] = 100

    db = gd.sageodata()

    if not resource:
        if report:
            logger.info(f"Finding resources for report: {report}")
            report = tn.Report(report, reporting_period)
            df = report.read_table("Report_Resources_mapping")
            resource_keys = [r for r in df[df.param == "WL"].resource_key.unique() if r]
            logger.info(f"Found resources: {resource_keys}")
    else:
        resource_keys = [resource]

    for resource in resource_keys:
        logger.info(
            f"Loading resource {resource} for reporting_period {reporting_period}"
        )
        resource = tn.Resource(resource, reporting_period)

        r = {}
        if step1:
            step1_data_validation(resource, db, plot_figures=figures)
        if step2:
            step2_defining_recovered_wls(resource, db, plot_figures=figures)
        if step3:
            step3_waterlevel_trends(resource, db, plot_figures=figures)

        if static:
            create_static_waterlevels_page(resource, db)


def step1_data_validation(resource, db, plot_figures=True):
    aq = gd.DEWAquarius()

    resource_key = resource.resource_key
    logger.info(f"Finding wells for {resource_key}...")
    wells = resource.find_wells()
    include_replacements = resource.well_selection_query.include_replacements

    logger.info(f"{len(wells)} wells with data from at least the reporting period")
    logger.debug(f" wells found: {wells.well_id.tolist()}")

    logger.info(f"Including replaced drillholes? {include_replacements}")

    logger.info("Retrieving manual water level data...")
    # wls = gd.fetch_wl_data(wells, include_replacements=include_replacements)
    sag_wl = db.water_levels(wells)
    sag_wl["database"] = "SA Geodata"
    sag_wl.loc[sag_wl.anomalous_ind == "Y", "grade"] = 15
    sag_wl.loc[sag_wl.anomalous_ind == "N", "grade"] = 30

    ## Doesn't work without access to SWIMS_Metadata
    logger.info("Identifying wells with logger data from SWIMSMetadata...")
    swimsmd = gd.SWIMSMetadata()
    dsets = swimsmd.datasets(wells.dh_no)
    aq_unit_hyphens = dsets.unit_hyphen.unique()
    # aq_unit_hyphens = sag_wl.unit_hyphen.unique()

    logger.info(f"Downloading logger data for {len(aq_unit_hyphens)} wells...")
    all_wl_dfs = [sag_wl]
    for unit_hyphen in aq_unit_hyphens:
        param = "rswl"
        row = wells[wells.unit_hyphen == unit_hyphen].iloc[0]
        logger.info(
            f" obtaining logger data at 1 day frequency for {unit_hyphen} / {row.obs_no}"
        )

        label = "Best Available"
        label_startswith = True
        unit_dsets = dsets[(dsets.unit_hyphen == unit_hyphen) & (dsets.sagd_param == param) & (dsets.label.str.contains("Best Available"))]
        if len(unit_dsets) > 1:
            unit_dsets["dset_length"] = unit_dsets.apply(
                lambda row: pd.Timestamp(row.last_data_timestamp) - pd.Timestamp(row.first_data_timestamp),
                axis=1
            )
            unit_dsets = unit_dsets.sort_values("dset_length", ascending=False)
            logger.warning(f"  {len(unit_dsets)} {param} BA datasets found?\n{unit_dsets[['label', 'first_data_timestamp', 'last_data_timestamp', 'dset_length']]}")
            logger.warning(f"  Selecting the first as it is the longest. Please fix in AQTS.")
            label = unit_dsets.iloc[0].label
            label_startswith = False

        aq_wl_dfs = aq.fetch_timeseries_data(
            row.unit_hyphen,
            param=param,
            label=label,
            label_startswith=label_startswith,
            freq="1d",
            max_gap_days=550,
            keep_grades=(0, 1, 5, 20, 30),
        )
        aq_wl = gd.join_logger_data_intervals(aq_wl_dfs)
        aq_wl = aq_wl.rename(
            columns={
                "timestamp": "obs_date",
            }
        )
        keep_cols = [
            "well_id",
            "obs_no",
            "dh_no",
            "unit_long",
            "unit_hyphen",
            "dh_name",
            "aquifer",
        ]
        for col in keep_cols:
            aq_wl[col] = row[col]
        aq_wl["database"] = "Aquarius"
        aq_wl["anomalous_ind"] = "N"
        aq_wl.loc[(aq_wl.grade <= 0) | (aq_wl.grade == 15), "anomalous_ind"] = "Y"
        if len(aq_wl):
            aq_wl["obs_date"] = aq_wl.obs_date.dt.tz_localize(None)
        keep_cols += ["database", "obs_date", "rswl", "anomalous_ind", "grade"]
        aq_wl = aq_wl[[c for c in keep_cols if c in aq_wl.columns]]
        all_wl_dfs.append(aq_wl)
    wls = pd.concat(all_wl_dfs).sort_values("obs_date")

    wls = wls.dropna(subset=["dtw", "swl", "rswl"], how="all")
    logger.debug(f"number of water level observations: {len(wls)}")

    logger.info("Filtering water level observations...")
    wls, removals = tn.filter_wl_observations(wls, return_removals=True, qc=resource)
    logger.debug(
        f"number of water level observations: retained = {len(wls)}; removed = {len(removals)}"
    )

    logger.info("Removing wells with fewer than 5 measurements...")
    wls = wls.groupby("well_id").filter(lambda x: len(x) >= 5)
    # removals = removals[removals.well_id.isin(wls.well_id.unique())]

    logger.info("Creating water level data validation graphs...")
    wls["well_title"] = wls.groupby("well_id").apply(
        lambda f: (f"{f.unit_hyphen.iloc[0]} {f.obs_no.iloc[0]} {f.dh_name.iloc[0]}")
    )
    removals["well_title"] = removals.groupby("well_id").apply(
        lambda f: (f"{f.unit_hyphen.iloc[0]} {f.obs_no.iloc[0]} {f.dh_name.iloc[0]}")
    )

    if resource_key.startswith("Far_North"):
        show_comments = False
    else:
        show_comments = True

    if plot_figures:
        for fn in resource.static_path.glob("plot_wl_data_validation_*.png"):
            logger.debug(f"removing old figure: {fn}")
            os.remove(fn)
        axes = tn.plot_wl_data_validation(
            wls,
            removals,
            well_title_col="well_title",
            show_comments=show_comments,
            adjust_comments=False,
            path=resource.static_path,
            savefig_and_close=True,
            dpi=100,
        )

    output_data = {"valid_data": wls, "invalid_data": removals}
    resource.write_data_sheets("validated_data", output_data)
    return output_data


def step2_defining_recovered_wls(resource, db, plot_figures=True):
    resource_key = resource.resource_key
    wls = resource.read_data("validated_data", sheet_name="valid data")

    wells = db.drillhole_details(wls.dh_no.unique())

    def analyse_wls_by_season(frame):
        well_id = frame.well_id.unique()[0]
        season = resource.get_season(well_id)
        return tn.analyse_wl_by_seasons(frame, season)

    logger.info("Calculating annual recovered water levels...")
    wlann = wls.groupby("well_id").apply(analyse_wls_by_season).reset_index(drop=True)

    # Create a subset of rows with only the recovery water levels.
    wlrec = wlann.loc[wlann.season == "recovery", :]
    logger.info(
        f"Filtering to only those with season == 'recovery' (retained records include wells {wlrec.well_id.unique()})"
    )

    # Remove any recovered levels which are beyond end_recovery_season
    end_y = int(str(resource.trend_dfn.end_recovery_season)[:4])
    wlrec = wlrec[wlrec.season_year.astype(str).str[:4].astype(int) <= end_y]
    logger.debug(
        f"Removed records >= end_recovery season ({resource.trend_dfn.end_recovery_season} -> {end_y}) by looking at column 'season_year'. Max date of remaining data = {wlrec.obs_date.max()}"
    )

    logger.info("Calculating the quality of record for each well...")
    qc, cmp_df = resource.get_wl_ranking_qc_results(wlrec)
    qc_results = pd.merge(
        qc, cmp_df["all_conditions"], left_index=True, right_index=True
    )
    for idx, well_qc_result in qc_results.iterrows():
        logger.info(
            f"\nQC results for {idx}\n" + pprint.pformat(well_qc_result.to_dict())
        )

    vals = resource.read_table("Data_validation")
    incl = vals.query("action == 'Include well in water level rankings'")
    for ix, row in incl.iterrows():
        for well_id in row.well_id.split(","):
            well_id = well_id.strip()
            if well_id in qc_results.index:
                x = qc_results.loc[well_id, "all_conditions"]
                qc_results.loc[well_id, "all_conditions"] = True
                comment = (
                    f"Included True by {row.username}: {row.comment} (overriding {x})"
                )
                qc_results.loc[well_id, "comment"] = comment
                logger.info(f"Data validation. {well_id} include=True {comment}")

    logger.debug(f"wlrec length={len(wlrec)}. Wells: {wlrec.well_id.unique()}")
    wlrec_ranked = wlrec.groupby("well_id", as_index=True).apply(
        lambda df: tn.rank_and_classify(
            df, param_name="rswl", param_col="rswl", dt_col="year+season"
        )
    )
    logger.debug(
        f"wlrec_ranked length={len(wlrec_ranked)}. Columns: {wlrec_ranked.columns}. Index values: \n{wlrec_ranked.index}"
    )
    wlrec_ranked = wlrec_ranked.reset_index(drop=True, level=1)
    wlrec_ranked = wlrec_ranked.drop(["rswl"], axis=1)
    logger.debug(
        f"wlrec_ranked  v2 length={len(wlrec_ranked)}. Columns: {wlrec_ranked.columns}. Index values: \n{wlrec_ranked.index}"
    )

    logger.info("Ranking and classifying recovered water levels for each well...")
    wlranks = pd.merge(
        wlrec.reset_index(),
        wlrec_ranked.reset_index(),
        on=("well_id", "year+season"),
    )
    logger.debug(f"wlranks wells: {wlranks.well_id.unique()}")

    end_year = int(resource.trend_dfn.end_year)
    next_year = end_year + 1
    next_year2 = str(int(next_year))[2:]
    end_year2 = f"{end_year}-{next_year2}"
    curr_year_recovery_seasons = [f"{end_year}-recovery", f"{end_year2}-recovery"]
    logger.info(f"Current year recovery seasons: {curr_year_recovery_seasons}")
    wlranks_curr = wlranks[wlranks["year+season"].isin(curr_year_recovery_seasons)]
    data_quality = (
        qc_results.reset_index()
    )  # pd.merge(qc, cmp_df, left_index=True, right_index=True).reset_index()
    logger.debug(f"data_quality wells: {data_quality.well_id.unique()}")

    wlrec = wlrec.reset_index()
    wlranks = (
        pd.merge(wlranks, data_quality[["well_id", "all_conditions"]], on="well_id")
        .rename(columns={"all_conditions": "meets_qc_filter"})
        .reset_index()
    )
    wlranks_curr = (
        pd.merge(
            wlranks_curr, data_quality[["well_id", "all_conditions"]], on="well_id"
        )
        .rename(columns={"all_conditions": "meets_qc_filter"})
        .reset_index()
    )

    wlranks_included = wlranks[wlranks.meets_qc_filter == True]
    wlranks_excluded = wlranks[wlranks.meets_qc_filter == False]
    wlranks_curr_included = wlranks_curr[wlranks_curr.meets_qc_filter == True]
    wlranks_curr_excluded = wlranks_curr[wlranks_curr.meets_qc_filter == False]

    logger.debug(f"wlranks_included: {wlranks_included.well_id.unique()}")
    logger.debug(f"wlranks_excluded: {wlranks_excluded.well_id.unique()}")
    logger.debug(f"wlranks_curr_included: {wlranks_curr_included.well_id.unique()}")
    logger.debug(f"wlranks_curr_excluded: {wlranks_curr_excluded.well_id.unique()}")

    logger.info("Calculate total water level changes.")
    windows = tn.get_total_wl_change_windows(current_year=end_year)
    total_wl_changes = tn.get_total_wl_changes(wlranks, windows=windows).reset_index()

    if plot_figures:
        for fn in resource.static_path.glob("plot_wl_bom_classes_*.png"):
            logger.debug(f"removing old figure: {fn}")
            os.remove(fn)
        for fn in resource.static_path.glob("plot_wl_seasonality_*.png"):
            logger.debug(f"removing old figure: {fn}")
            os.remove(fn)

        logger.info(
            "Chart the original (filtered) water levels against the ranked and classified recovered water levels..."
        )
        for well_id, wwlranks in wlranks.groupby(wlranks.well_id):
            logger.info(f"{well_id}...")
            r = tn.plot_wl_bom_classes(
                wwlranks, wls[wls.well_id == well_id], fig=plt.figure(figsize=(10, 4))
            )
            titledf = wls[wls.well_id == well_id].iloc[0]
            well_title = f"{titledf.unit_hyphen} {titledf.obs_no} {titledf.dh_name}"
            r["ax"].set_title(
                f"{well_title}\nmeets all_conditions {cmp_df.loc[well_id].all_conditions}"
            )
            r["ax"].figure.tight_layout()
            filename = str(resource.static_path / f"plot_wl_bom_classes_{well_id}.png")
            logger.debug(f"filename = {filename}")
            r["ax"].figure.savefig(filename, dpi=100)
            tn.trim_whitespace_from_image(filename, append=False)

        logger.info("Chart seasonality for each well...")
        for well_id, wwlann in wlann.groupby("well_id"):
            logger.info(f"Plotting seasonal chart for {well_id}")
            seasons = resource.get_season(well_id)
            r = tn.plot_wl_seasonality(
                wls[wls.well_id == well_id], wwlann, seasons=seasons, dots=True
            )
            titledf = wls[wls.well_id == well_id].iloc[0]
            well_title = f"{titledf.unit_hyphen} {titledf.obs_no} {titledf.dh_name}"
            r["fig"].suptitle(well_title)
            filename = str(resource.static_path / f"plot_wl_seasonality_{well_id}.png")
            r["fig"].savefig(filename, dpi=100)
            tn.trim_whitespace_from_image(filename, append=False)

        logger.info("Charting: plot_wl_rankings_internal")
        fig = tn.plot_wl_rankings_internal(wlranks_curr, title=resource.resource_key)
        filename = f"{resource.resource_key}_wl_rankings_internal.png"
        for path in [resource.data_path]:
            fn = str(path / filename)
            fig.savefig(fn, dpi=120)
            tn.trim_whitespace_from_image(fn, append=False)

        logger.info("Charting: plot_wl_historical_rankings")
        fig = tn.plot_wl_historical_rankings(
            wlranks, data_quality, title=resource.resource_key
        )
        filename = f"{resource.resource_key}_wl_historical_rankings.png"
        for path in [resource.data_path]:
            fn = str(path / filename)
            fig.savefig(fn, dpi=120)
            tn.trim_whitespace_from_image(fn, append=False)

    data_quality = data_quality.reset_index()

    output_data = {
        "annual_recovered_wl": wlrec,
        "ranked_wls": wlranks_included,
        "current_ranked_wls": wlranks_curr_included,
        "ranks_excl": wlranks_excluded,
        "current_ranks_excl": wlranks_curr_excluded,
        "total_wl_changes": total_wl_changes,
        "data_quality": data_quality,
    }
    resource.write_data_sheets("recovery_wl_data", output_data)

    for path in [resource.data_path]:
        spath = path / "spatial_outputs"
        spath.mkdir(exist_ok=True)
        tn.df_to_shp(
            wlranks, spath / f"ranked_recovered_wls_{resource.resource_key}.shp"
        )
        tn.df_to_shp(
            wlranks_curr,
            spath / f"ranked_current_recovered_wls_{resource.resource_key}.shp",
        )

    return output_data


def step3_waterlevel_trends(resource, db, plot_figures=True):
    resource_key = resource.resource_key

    wls = resource.read_data("validated_data", sheet_name="valid data")
    if resource.trend_dfn.regress_against.startswith("all"):
        logger.info("Reading all valid data for trend regression.")
        wldata = pd.DataFrame(wls)
        wldata["season_year"] = wldata.obs_date.dt.year
    elif resource.trend_dfn.regress_against.startswith("recovery"):
        logger.info("Reading all recovery seasonal WL data for trend regression.")
        wldata = resource.read_data(
            "recovery_wl_data", sheet_name="annual recovered WL"
        )
    else:
        raise KeyError(
            "Definitions_triclass_trend_thresholds:regress_against must be either 'recovery' or 'all'"
        )
    if "SA Geodata only" in resource.trend_dfn.regress_against:
        wldata = wldata[wldata.database == "SA Geodata"]

    wldata["ndays"] = wldata.obs_date.apply(tn.date_to_decimal)

    logger.info("Filtering data tables to the trend period")
    wldata = wldata.groupby("well_id").filter(resource.test_wls_in_trend_period)

    wells = db.drillhole_details(wldata.dh_no.unique())

    trenddf = (
        wldata.groupby("well_id", as_index=False)
        .apply(lambda wdf: resource.filter_wl_to_trend_period(wdf))
        .reset_index(drop=True)
    )

    logger.info(
        "Calculate how many observations fall within the trend period and exclude any if needed."
    )
    vals = resource.read_table("Data_validation")
    excl = vals[(vals.action == "Exclude well trend") | (vals.action == "Exclude well")]
    incl = vals.query("action == 'Include well trend'")
    trendqc = trenddf.groupby("well_id").rswl.count().to_frame(name="n_trend_obs")
    trendqc["include"] = True
    trendqc["exclusion_reason"] = ""

    for well_id in trendqc[trendqc.n_trend_obs < resource.trend_dfn.min_data_pts].index:
        trendqc.loc[well_id, "include"] = False
        trendqc.loc[
            well_id, "exclusion_reason"
        ] = f"Only {trendqc.loc[well_id].n_trend_obs} data pts in trend period"

    for ix, row in excl.iterrows():
        for well_id in row.well_id.split(","):
            well_id = well_id.strip()
            if well_id in trendqc.index:
                trendqc.loc[well_id, "include"] = False
                trendqc.loc[
                    well_id, "exclusion_reason"
                ] = f"{row.username}: {row.comment}"
    for ix, row in incl.iterrows():
        for well_id in row.well_id.split(","):
            well_id = well_id.strip()
            if well_id in trendqc.index:
                trendqc.loc[well_id, "include"] = True
                x = trendqc.loc[well_id, "exclusion_reason"]
                if x:
                    x = f" (overriding '{x}')"
                trendqc.loc[well_id, "exclusion_reason"] = (
                    f"{row.username}: {row.comment}" + x
                )
    logger.info(str(trendqc))

    logger.info("Calcuate trend line by linear regression...")
    welltrends = trenddf.groupby("well_id").apply(lambda wdf: resource.apply_trend(wdf))

    if plot_figures:
        for fn in resource.static_path.glob("plot_wl_trend_*.png"):
            logger.debug(f"removing old figure: {fn}")
            os.remove(fn)

        trend_line_start = pd.Timestamp(
            f"{int(str(resource.trend_dfn.start_recovery_season)[:4])}-01-01"
        )
        trend_line_end = pd.Timestamp(
            f"{int(str(resource.trend_dfn.end_recovery_season)[:4])}-12-31"
        )

        logger.info("Chart all trend lines...")
        for well_id in sorted(wells.well_id):
            titledf = wldata[wldata.well_id == well_id].iloc[0]
            well_title = f"{titledf.unit_hyphen} {titledf.obs_no} {titledf.dh_name}"
            wldata.loc[wldata.well_id == well_id, "well_title"] = well_title
            logger.info(f"{well_id} ({well_title})...")
            ax = tn.plot_wl_trend(
                wldata.query(f'well_id == "{well_id}"'),
                trenddf.query(f'well_id == "{well_id}"'),
                well_title_col="well_title",
                all_wls=wls.query(f'well_id == "{well_id}"'),
                trend_lines=welltrends,
                trend_line_start=trend_line_start,
                trend_line_end=trend_line_end,
                override_year_span=False,
            )
            tt = ax._title_text.get_text()
            exclusion_label = "\n".join(
                textwrap.wrap(trendqc.loc[well_id].exclusion_reason, 50)
            )
            if not trendqc.loc[well_id].include:
                tt += f"\nExcluded: {exclusion_label}"
            else:
                if exclusion_label:
                    tt += f"\nIncluded: {exclusion_label}"
                else:
                    tt += f" - Included"
            ax._title_text.set_text(tt)
            ax.figure.tight_layout()
            filename = str(resource.static_path / f"plot_wl_trend_{well_id}.png")
            ax.figure.savefig(filename, dpi=100)
            tn.trim_whitespace_from_image(filename, append=False)

    logger.info("Merge trend results into one table...")
    trend_results = pd.merge(
        welltrends,
        trendqc,
        left_on="well_id",
        right_index=True,
    )
    trend_results = trend_results[
        [x for x in trend_results.columns if not x.startswith("Unnamed")]
    ]

    curr_ranks = resource.read_data("recovery_wl_data", "current ranked WLs")

    cols = [
        "well_id",
        "dh_no",
        "unit_long",
        "unit_hyphen",
        "obs_no",
        "dh_name",
        "easting",
        "northing",
        "zone",
        "latitude",
        "longitude",
        "aquifer",
    ]
    trend_results = pd.merge(
        wldata[cols].drop_duplicates(),
        trend_results,
        on="well_id",
        how="inner",
    )

    trend_results["status_combined"] = trend_results["status_change"]
    for well_id in curr_ranks.well_id:
        well_idx = trend_results.well_id == well_id
        if len(well_idx[well_idx == True]) > 0:
            trend_status = trend_results.loc[well_idx, "status_change"].iloc[0]
            rank_status = curr_ranks.loc[
                curr_ranks.well_id == well_id, "rswl_bom_class"
            ].iloc[0]
            trend_results.loc[
                well_idx, "status_combined"
            ] = f"{trend_status} ({rank_status})"
        else:
            trend_results["status_combined"] = trend_results["status_change"]

    trend_results.loc[
        trend_results["status_change"] == "Rising", "gsr_status_colour"
    ] = "Blue"
    trend_results.loc[
        trend_results["status_change"] == "Stable", "gsr_status_colour"
    ] = "Green"
    trend_results.loc[
        trend_results["status_change"] == "Declining", "gsr_status_colour"
    ] = "Yellow"
    trend_results.loc[
        trend_results["status_combined"] == "Rising (Lowest on record)",
        "gsr_status_colour",
    ] = "Purple"
    trend_results.loc[
        trend_results["status_combined"] == "Stable (Lowest on record)",
        "gsr_status_colour",
    ] = "Grey"
    trend_results.loc[
        trend_results["status_combined"] == "Declining (Lowest on record)",
        "gsr_status_colour",
    ] = "Black"

    wldata = wldata.reset_index()
    trenddf = trenddf.reset_index()
    final_trends = trend_results[trend_results.include == True].reset_index()
    excluded_trends = trend_results[trend_results.include == False].reset_index()
    trendqc = trendqc.reset_index()

    output_data = {
        "wl_data": wldata,
        "wls_in_trend_period": trenddf,
        "trend_qc": trendqc,
        "final_trends": final_trends,
        "excluded_trends": excluded_trends,
    }
    resource.write_data_sheets("recovery_wl_trends", output_data)

    for path in [resource.data_path]:
        spath = path / "spatial_outputs"
        spath.mkdir(exist_ok=True)
        tn.df_to_shp(
            final_trends, spath / f"recovered_wl_trends_{resource.resource_key}.shp"
        )

    return output_data


def create_static_waterlevels_page(resource, db):
    templates = tn.load_html_templates()
    template = templates.get_template("static/waterlevels.html")

    retdata = tn.collate_waterlevel_summary_data(resource, db)
    sentences = tn.construct_waterlevel_template_sentences(resource, retdata)
    retdata["sentences"] = sentences

    with open(str(resource.static_path / "waterlevels.html"), mode="w") as html:
        html.write(template.render(retdata))


if __name__ == "__main__":
    run()



import colorsys
from datetime import date, datetime, timedelta
import io
import re
import os
import logging
import textwrap
from pathlib import Path

import matplotlib.pyplot as plt
from matplotlib import lines as mlines
from matplotlib import gridspec
from matplotlib import dates as mdates
from matplotlib import ticker as mticker
from matplotlib import colors as mcolors
import numpy as np
import pandas as pd
from adjustText import adjust_text
from scipy import stats
from PIL import Image, ImageChops
import pyproj

from .gwutils import *
from .utils import *
from .rainfall import *
from .charts_utils import *

logger = get_logger()


def plot_rainfall_wuy_annual(
    rf, ax=None, trend_line=False, rf_col="Rain", dt_col="Date", wuy_col=False
):
    """Plot annual rainfall.

    Args:
        rf (pandas DataFrame): daily rainfall data
        trend_line (bool): calculate and plot a trend line
        rf_col (str): column containing rainfall in mm
        dt_col (str): column containing datetimes
        wuy_col (str or False): column containing the water-use year. If it is
            False then the column will be calculated and added as "wu_year"

    Note, water-use year is the same as financial year.

    Returns:
        dict with keys "ax", "rf_wuy" (a pandas Series), "slope" (mm/y) and
        "intercept".

    """

    width = 0.8
    if ax is None:
        fig = plt.figure(figsize=(7.2, 3.4))
        ax = fig.add_subplot(111)

    if not wuy_col in rf:
        rf.loc[:, "wu_year"] = [date_to_wateruseyear(d) for d in rf[dt_col]]
    rf_wuy = rf.groupby("wu_year")[rf_col].sum()
    rf_wuy_x = np.arange(len(rf_wuy))
    slope, intercept, r_value, p_value, std_err = stats.linregress(
        rf_wuy_x, rf_wuy.values
    )

    ax.bar(
        rf_wuy_x,
        rf_wuy.values,
        width=width,
        facecolor=rainfall_colours["rainfall"],
        label="Annual rainfall",
    )

    trend_x = np.asarray([rf_wuy_x[0], rf_wuy_x[-1]])
    trend_y = [slope * tx + intercept for tx in trend_x]
    if trend_line:
        ax.plot(trend_x, trend_y, color="darkblue", label="Trend in rainfall")

    ax.plot(
        [rf_wuy_x[0] - width / 2, rf_wuy_x[-1] + width / 2],
        [rf_wuy.mean(), rf_wuy.mean()],
        color=rainfall_colours["mean"],
        ls="--",
        lw=1,
        label=f"Mean annual rainfall ({rf_wuy.mean():.0f} mm)",
    )

    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)

    ticks_per_label = 2
    xticks = np.arange(0, len(rf_wuy_x), ticks_per_label)
    while xticks[-1] != rf_wuy_x[-1]:
        xticks += 1
    ax.set_xticks(xticks)

    _ = ax.set_xticklabels(
        rf_wuy.index[xticks], rotation=90, ha="center", fontsize="medium"
    )
    y0, y1 = ax.get_ylim()
    y1_new = y1 + (y1 - y0) / 5
    ax.set_ylim(y0, y1_new)
    ax.set_ylabel("Annual rainfall (mm)", fontsize="medium")
    ax.legend(loc="best", fontsize="small", frameon=False, ncol=2)

    chart = {"ax": ax, "rf_wuy": rf_wuy, "slope": slope, "intercept": intercept}
    return chart


def plot_rainfall_calendar_annual(
    rf, ax=None, trend_line=False, rf_col="Rain", dt_col="Date", year_col=False
):
    """Plot annual rainfall.

    Args:
        rf (pandas DataFrame): daily rainfall data
        trend_line (bool): calculate and plot a trend line
        rf_col (str): column containing rainfall in mm
        dt_col (str): column containing datetimes
        year_col (str or False): column containing the year as an
            integer. If it is
            False then the column will be calculated and added as "year"

    Returns:
        dict with keys "ax", "rf_year" (a pandas Series), "slope" (mm/y) and
        "intercept".

    """

    width = 0.8
    if ax is None:
        fig = plt.figure(figsize=(7.2, 3.4))
        ax = fig.add_subplot(111)

    if not year_col in rf:
        rf.loc[:, "year"] = rf[dt_col].dt.year
    rf_year = rf.groupby("year")[rf_col].sum()
    rf_year_x = np.arange(len(rf_year))
    slope, intercept, r_value, p_value, std_err = stats.linregress(
        rf_year_x, rf_year.values
    )

    ax.bar(
        rf_year_x,
        rf_year.values,
        width=width,
        facecolor=rainfall_colours["rainfall"],
        label="Annual rainfall",
    )

    trend_x = np.asarray([rf_year_x[0], rf_year_x[-1]])
    trend_y = [slope * tx + intercept for tx in trend_x]
    if trend_line:
        ax.plot(trend_x, trend_y, color="darkblue", label="Trend in rainfall")

    ax.plot(
        [rf_year_x[0] - width / 2, rf_year_x[-1] + width / 2],
        [rf_year.mean(), rf_year.mean()],
        color=rainfall_colours["mean"],
        ls="--",
        lw=1,
        label=f"Mean annual rainfall ({rf_year.mean():.0f} mm)",
    )

    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)

    ticks_per_label = 2
    xticks = np.arange(0, len(rf_year_x), ticks_per_label)
    while xticks[-1] != rf_year_x[-1]:
        xticks += 1
    ax.set_xticks(xticks)

    _ = ax.set_xticklabels(
        rf_year.index[xticks], rotation=90, ha="center", fontsize="medium"
    )
    y0, y1 = ax.get_ylim()
    y1_new = y1 + (y1 - y0) / 5
    ax.set_ylim(y0, y1_new)
    ax.set_ylabel("Annual rainfall (mm)", fontsize="medium")
    ax.legend(loc="best", fontsize="small", frameon=False, ncol=2)

    chart = {"ax": ax, "rf_year": rf_year, "slope": slope, "intercept": intercept}
    return chart


def plot_rainfall_monthly_means(
    mm,
    pm,
    mlabel,
    plabel,
    xlabel_left,
    xlabel_right,
    ax=None,
    xtl_kws=None,
    year_type="split",
    yl_fontsize="medium",
    leg_fontsize="small",
    xl2_fontsize="medium",
    bar_width=0.38,
    y_bump=0.1,
    ly=0.02,
):
    """Plot monthly totals vs monthly means.

    Args:
        mm (pandas Series): index is month number, value is
            monthly mean rainfall for that month
        pm (pandas Series): multilevel index: year, month number: the
            values are the monthly totals for that month.
        mlabel (str): label for the mean period e.g. "1977-78 to 2018-19"
        plabel (str): label for period e.g. "2018-19"
        year_type (str): either "split" (water use year) or "calendar"
        # # xlabel_left (str): label for left side of x-axis e.g. "2018"
        # # xlabel_right (str): label for right side of x-axis e.g. "2019"
        # xtl_fontsize (str): xticklabel fontsize
        y_bump (float): figure fraction to shift x-axis up by

    Returns: ax

    """

    width = bar_width
    month_numbers = pm.index.get_level_values(1)
    month_names = [date(2000, mn, 1).strftime("%b") for mn in month_numbers]
    rf_period_mm = [mm.loc[mn] for mn in month_numbers]
    rf_period_m = [x for x in pm]

    if ax is None:
        fig = plt.figure(figsize=(7.2, 3.0))
        ax = fig.add_subplot(111)
    rf_period_x = np.arange(len(month_numbers)) - width / 2
    rf_period_mm_x = np.arange(len(month_numbers)) + width / 2
    ax.bar(
        rf_period_x,
        rf_period_m,
        width=width,
        facecolor=rainfall_colours["rainfall"],
        label=f"{plabel} rainfall",
    )
    ax.bar(
        rf_period_mm_x,
        rf_period_mm,
        width=width,
        facecolor=rainfall_colours["mean"],
        label=f"Mean rainfall\n({mlabel})",
    )
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.set_xticks(range(len(month_numbers)))
    if xtl_kws is None:
        xtl_kws = dict(rotation=0, ha="center", fontsize="medium")
    _ = ax.set_xticklabels(month_names, **xtl_kws)
    ax.set_ylabel("Monthly rainfall (mm)", fontsize=yl_fontsize)
    ax.legend(loc="upper right", fontsize=leg_fontsize, frameon=False, ncol=2)
    y0, y1 = ax.get_ylim()
    y1_new = y1 + (y1 - y0) / 6
    ax.set_ylim(y0, y1_new)
    if year_type == "split":
        pos = ax.get_position()
        ax.set_position([pos.x0, pos.y0 + y_bump, pos.width, pos.height - y_bump])

        # ly = 0.02
        l1_x = (pos.x0, pos.x0 + 0.31)
        l2_x = [pos.x0 + 0.34, pos.x0 + pos.width + 0.04]
        ax.text(
            l1_x[0] + (l1_x[1] - l1_x[0]) / 2,
            ly,
            xlabel_left,
            transform=ax.figure.transFigure,
            fontsize=xl2_fontsize,
            va="center",
            ha="center",
            bbox=dict(facecolor="white", edgecolor="none"),
        )
        ax.text(
            l2_x[0] + (l2_x[1] - l2_x[0]) / 2,
            ly,
            xlabel_right,
            transform=ax.figure.transFigure,
            fontsize=xl2_fontsize,
            va="center",
            ha="center",
            bbox=dict(facecolor="white", edgecolor="none"),
        )

        line1 = mlines.Line2D(
            l1_x, [ly, ly], color="k", lw=0.5, transform=ax.figure.transFigure
        )
        line2 = mlines.Line2D(
            l2_x, [ly, ly], color="k", lw=0.5, transform=ax.figure.transFigure
        )
        for line in (line1, line2):
            line.set_clip_on(False)
            ax.add_line(line)
    return ax


def plot_monthly_rainfall_internal(
    station,
    report,
    show_year=None,
    dailydf=None,
    chart_start=2000,
    stat_start=None,
    stat_end=None,
):
    """Plot monthly average rainfall against a given year (either calendar or
    financial) and show the amount of change in labels on the chart.

    Args:
        station (dict): station to analyse. Must have keys "dfn" (pd.Series),
            and "station" (str). See :meth:`wrap_technote.Report.find_rainfall_station`.
        report (:class:`wrap_technote.Report`): report definition
            which defines the mean rainfall period and access to data.
        show_year (str or None): the year which should be shown in comparison to the
            mean period. By default this will be the most recent year in the data,
            so if you run the code in Jan or Feb it may be mostly blank ;-)
            If you specify a calendar year e.g. "2020" then the chart will
            extend from January on the left to December on the right.
            If you specify a financial year e.g. "2019-20" then the chart will
            extend from July on the left to June on the right.
        dailydf (pd.DataFrame): the data to plot. If not specified, daily
            data for *station* will be retrieved via *report*.
        chart_start (str/int): start year for the chart display only, can be
            either e.g. "2000" or 2000, or financial year e.g. "1998-99".
        stat_start, stat_end (str/int): as for *chart_start*, defines the
            mean period. If not supplied, it will be taken from
            :meth:`wrap_technote.Report.rainfall_dfn`.

    Example code::

        >>> import wrap_technote as tn
        >>> report = tn.load_report("Eyre Pen", "2020-21")
        >>> station = report.find_rainfall_station(id='18069')
        >>> tn.plot_monthly_rainfall_internal(station, report, show_year="2020")

    .. figure:: figures/monthly_rainfall_internal_2020_18069_ELLISTON.png

    You may need a higher DPI to see the smaller labels (`plt.rcParams['figure.dpi'] = 100`).

    Returns: matplotlib Axes object.

    """

    ddf = dailydf
    if stat_start is None:
        stat_start = station["dfn"].mean_period_start
    if stat_end is None:
        stat_end = station["dfn"].mean_period_finish - pd.Timedelta(days=365)

    if ddf is None:
        ddf = report.get_rainfall_data(station, "daily")

    ddf["year"] = ddf["Date"].dt.year

    if show_year is None:
        show_year = str(ddf.year.max())

    if "-" in show_year:
        year_col = "wu_year"
        dt_col = "wu_year"
    else:
        year_col = "year"
        dt_col = "year"

    monthly = reduce_daily_to_monthly(ddf, year_col=year_col)

    stat_start = stat_start.year
    stat_end = stat_end.year
    title = f"{report.report_key} - monthly - mean from {stat_start} - {station['station']} ({station['dfn'].station_name})"

    chart_idx = lambda x: x[dt_col].astype(str).str[:4].astype(int) >= chart_start
    stat_idx = lambda x: (x[dt_col].astype(str).str[:4].astype(int) >= stat_start) & (
        x[dt_col].astype(str).str[:4].astype(int) < (stat_end + 1)
    )
    stat_0 = sorted(ddf.loc[stat_idx, dt_col].values)[0]
    stat_1 = sorted(ddf.loc[stat_idx, dt_col].values)[-1]

    select_idx = lambda x: (x[year_col].astype(str) == show_year)

    stat_months = monthly[stat_idx].groupby("month").mean()
    select_months = monthly[select_idx].groupby("month").mean()
    # print(stat_months)

    fig = plt.figure(figsize=(8.8, 4.5))
    ax = fig.add_subplot(111)
    if dt_col == "wu_year":
        leg_loc = "upper center"
        ax_idx = np.array([7, 8, 9, 10, 11, 12, 1, 2, 3, 4, 5, 6])
        year_for_month = (
            lambda m: int(show_year[:4]) if m >= 7 else int(show_year[:4]) + 1
        )
    elif dt_col == "year":
        leg_loc = "upper left"
        ax_idx = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
        year_for_month = lambda m: int(show_year)

    ax_months = [
        pd.Timestamp(year=year_for_month(m), month=m, day=1).strftime("%b\n%Y")
        for m in ax_idx
    ]
    # fill any missing data
    select_months = select_months.reindex(index=ax_idx)

    x_idx = np.arange(0, len(ax_idx))
    width = 0.42
    ax.bar(
        x_idx - width / 2,
        select_months.loc[ax_idx]["Rain"],
        width=width,
        fc="tab:blue",
        label=f"{show_year}",
        alpha=0.5,
    )
    ax.bar(
        x_idx + width / 2,
        stat_months.loc[ax_idx]["Rain"],
        width=width,
        fc="tab:green",
        label=f"Mean {stat_0} to {stat_1}",
        alpha=0.5,
    )
    ax.plot(
        x_idx - width / 6,
        select_months.loc[ax_idx]["Rain"],
        marker="o",
        c="tab:blue",
        label="",
    )
    ax.plot(
        x_idx + width / 6,
        stat_months.loc[ax_idx]["Rain"],
        marker="o",
        ls=":",
        c="tab:green",
        label="",
    )
    ax.axhline(0, color="k", lw=0.5)
    _ = ax.set_xticks(x_idx)
    ax.set_xticklabels(ax_months)

    ax.legend(loc="best", frameon=False, fontsize="small")
    y1 = ax.get_ylim()[1]
    yspan = y1 - ax.get_ylim()[0]
    box_space_height = yspan / 7
    box_step = box_space_height / 5
    box_height = box_space_height - box_step
    y0 = 0 - box_space_height

    for m_idx, m in enumerate(ax_idx):
        sel = select_months.loc[m].Rain
        stat = stat_months.loc[m].Rain
        diff = sel - stat
        diff_pct = diff / stat * 100
        cmp_pct = sel / stat * 100
        if diff > 0:
            box_y = 0
        else:
            box_y = 0 - box_step
        label = (
            f"{sel:.0f} ({stat:.0f})\n{diff:+.0f} ({diff_pct:+.0f}%)\n({cmp_pct:.0f}%)"
        )
        ax.text(
            m_idx,
            box_y - (box_height / 10),
            label,
            va="top",
            ha="center",
            fontsize="xx-small",
            bbox=dict(boxstyle="square", fc="white", ec="none", alpha=0.8),
        )
    ax.set_ylim(y0, y1)
    ax.set_ylabel("Monthly rainfall (mm)")
    ax.set_title(title)
    return ax


def plot_annual_rainfall_internal(
    station,
    report,
    df=None,
    chart_start=2000,
    stat_start=None,
    stat_end=None,
    sheet_name="annual (water-use year)",
    annotations=True,
    ax=None,
    fig=None,
    report_key="",
    fig_width=None,
):
    """Plot chart for internal analysis of annual rainfall.

    Args:
        station (dict): from :meth:`wrap_technote.Report.get_rainfall_station`.
            must have key "dfn" containing pd.Series with index values
            "mean_period_start" and "mean_period_finish"
        chart_start (int): year in which the chart begins
        stat_start (pd.Timestamp): date from which the mean begins (will be taken from
            *station* if None)
        stat_end (pd.Timestamp): date on which the mean ends (will be taken from
            *station* if None)
        sheet_name (str): method of defining a year, must be either 'financial' or 'calendar'
        annotations (bool): draw the labels
        report_key (str): "" - used only in the title
        ax (matplotlib Axes)
        fig (matplotlib Figure)

    """

    if "water" in sheet_name.lower() and "use" in sheet_name.lower():
        sheet_name = "annual (water-use year)"
    elif "fin" in sheet_name.lower():
        sheet_name = "annual (water-use year)"
    elif "calendar" in sheet_name.lower():
        sheet_name = "annual (calendar)"
    else:
        raise KeyError(
            f"sheet_name {sheet_name} must be either 'financial' or 'calendar'"
        )

    if df is None:
        logger.debug(f"Loading data for {station['name']}")
        df = report.get_rainfall_data(station, sheet_name)

    if sheet_name == "annual (water-use year)":
        dt_col = "wu_year"
        df = df[(df[dt_col].str[:4].astype(int) + 1) <= datetime.today().year]
    elif sheet_name == "annual (calendar)":
        dt_col = "Date"
        df = df[df[dt_col] < datetime.today().year]

    if stat_start is None:
        stat_start = station["dfn"].mean_period_start
    if stat_end is None:
        stat_end = station["dfn"].mean_period_finish

    stat_start = stat_start.year
    if sheet_name == "annual (water-use year)":
        stat_end = (stat_end - pd.Timedelta(days=365)).year
    else:
        stat_end = stat_end.year

    logger.debug(f"stat_start = {stat_start} and stat_end = {stat_end}")

    # title = f"{report.report_key} - {sheet_name} - mean from {stat_start} - {station['station']} ({station['dfn'].station_name})"
    title = f"{station['id']} {station['name']}"

    chart_idx = lambda x: x[dt_col].astype(str).str[:4].astype(int) >= chart_start
    wu_years = list(df[chart_idx][dt_col].unique())

    if fig_width is None:
        fig_width = len(wu_years) * 0.4 + 0.8

    chart_idx = lambda x: x[dt_col].astype(str).str[:4].astype(int) >= chart_start
    stat_idx = lambda x: (x[dt_col].astype(str).str[:4].astype(int) >= stat_start) & (
        x[dt_col].astype(str).str[:4].astype(int) < (stat_end + 1)
    )
    stat_0 = sorted(df.loc[stat_idx, dt_col].values)[0]
    stat_1 = sorted(df.loc[stat_idx, dt_col].values)[-1]
    mean = df[stat_idx].Rain.mean()
    df["rain_change"] = df.Rain.diff()
    df["rain_pct_change"] = df.Rain.pct_change() * 100
    df["mean_change"] = df.Rain - mean
    df["mean_pct_change"] = df.mean_change / mean * 100
    df2 = df[chart_idx]
    df2 = df2.set_index(dt_col).reindex(wu_years).reset_index()
    if fig is None:
        fig = plt.figure(figsize=(fig_width, 5))
    if ax is None:
        ax = fig.add_subplot(111)
    ax.bar(df2.index, df2.Rain)
    ax.set_xticks(df2.index)
    ax.set_xticklabels(df2[dt_col])
    plt.setp(ax.get_xticklabels(), rotation=50, ha="right")
    ax.axhline(mean, color="brown", lw=1)
    mean_labelled = False
    for ix, row in df2.iterrows():
        if row.rain_change > 0:
            boxy = 30
        else:
            boxy = 5
        ax.text(
            ix,
            #             row.Rain + row.Rain / 22,
            boxy,
            (
                f"{row.Rain:.0f}\n"
                + f"{row.rain_change:+.0f}\n"
                + f"{row.rain_pct_change:+.0f}%"
            ),
            va="bottom",
            ha="center",
            fontsize="xx-small",
            bbox=dict(boxstyle="square", fc="white", ec="none", alpha=0.8),
        )

        # Position of top or bottom of the comparison to the mean.
        y = row.Rain + (row.Rain / 20 * (1 if row.mean_change > 0 else -1))
        ax.annotate(
            "",
            (ix, y),
            (ix, mean),
            arrowprops=dict(arrowstyle="-|>", color="brown", lw=0.5),
        )
        # Comparison to the mean - text label.
        ax.text(
            ix,
            y,
            (f"{row.mean_change:+.0f}\n" + f"{row.mean_pct_change:+.0f}%"),
            va=("bottom" if row.mean_change > 0 else "top"),
            ha="center",
            fontsize="xx-small",
            color="brown",
            bbox=dict(boxstyle="square", fc="white", ec="none", alpha=0.8),
        )

    ax.set_title(title + f"\nmean {stat_0} to {stat_1}")
    ax.set_ylim(None, ax.get_ylim()[1] + df2.Rain.max() / 10)
    ax.set_xlim(df2.index[0] - 1.5, None)
    ax.text(
        df2.index[0] - 1.3,
        mean,
        f"{mean:.0f}:",
        ha="left",
        va="center",
        style="italic",
        fontsize="x-small",
        color="brown",
        bbox=dict(boxstyle="square", fc="white", alpha=0.8, ec="none"),
    )
    ax.text(
        1,
        1,
        f"Data ends {df[dt_col].max()}",
        transform=ax.transAxes,
        ha="right",
        va="top",
        fontsize="x-small",
    )
    return ax


def plot_seasonal_rainfall_residuals_internal(
    station,
    report,
    daily_df=None,
    chart_start=2000,
    chart_end=None,
    mean_start=None,
    mean_end=None,
    value_type="%",
    fig_width=None,
):
    if daily_df is None:
        ddf = report.get_rainfall_data(station, "daily")
    else:
        ddf = daily_df

    ddf["year"] = ddf.Date.dt.year
    monthly = reduce_daily_to_monthly(ddf, year_col="year")

    if chart_end is None:
        # chart_end = (station["dfn"].rf_period_finish - pd.Timedelta(days=365)).year
        chart_end = station["dfn"].rf_period_finish.year
    if mean_start is None:
        mean_start = station["dfn"].mean_period_start.year
    if mean_end is None:
        # mean_end = (station["dfn"].mean_period_finish - pd.Timedelta(days=365)).year
        mean_end = station["dfn"].rf_period_finish.year

    value_type_title = "pct" if value_type == "%" else value_type

    # title = f"{report.report_key} - seasonal difference {value_type_title} - mean from {mean_start} - {station['station']} ({station['dfn'].station_name})"
    title = f"{station['id']} {station['name']}"

    monthly["season"] = monthly.month.map(months_to_seasons)
    monthly.loc[monthly.season == "summer", "season"] = "0-summer"
    monthly.loc[monthly.season == "autumn", "season"] = "1-autumn"
    monthly.loc[monthly.season == "winter", "season"] = "2-winter"
    monthly.loc[monthly.season == "spring", "season"] = "3-spring"
    monthly["year-season"] = monthly["year"].astype(str) + "-" + monthly["season"]
    monthly = monthly.sort_values(["year", "month"])
    seasons = (
        monthly.groupby(["year-season", "year", "season"]).Rain.sum().reset_index()
    )

    smeans = (
        seasons[(seasons.year >= mean_start) & (seasons.year <= mean_end)]
        .groupby("season")
        .Rain.mean()
    )
    rows = []
    index = []
    for year, s in seasons.groupby("year"):
        year_totals = s.set_index("season").Rain
        year_diffs = year_totals - smeans
        index.append(year)
        rows.append(year_diffs)
    sdiffs = pd.DataFrame(rows, index=index)
    if value_type == "%":
        for col in sdiffs.columns:
            sdiffs[col] = sdiffs[col] / smeans.loc[col] * 100

    annrf = seasons.groupby("year").Rain.sum()
    annrf_mean = annrf[(annrf.index >= mean_start) & (annrf.index <= mean_end)].mean()
    annrf_dev = annrf - annrf_mean
    if value_type == "%":
        annrf_dev = annrf_dev / annrf_mean * 100

    if chart_end is None:
        chart_end = sdiffs.index.max()

    years = chart_end - chart_start + 1
    years_plot = np.array(
        [y for y in np.arange(chart_start, chart_end + 1, 1) if y in sdiffs.index]
    )
    sdiffs_plot = sdiffs[(sdiffs.index >= chart_start) & (sdiffs.index <= chart_end)]

    if fig_width is None:
        fig_width = 0.8 + len(years_plot) * 0.3
    fig = plt.figure(figsize=(fig_width, 5.5))
    ax = fig.add_subplot(111)
    season_names = ["0-summer", "1-autumn", "2-winter", "3-spring"]
    s_width = 0.25
    for i, season in enumerate(season_names):
        x_offset = i * s_width
        ax.bar(
            years_plot + x_offset + (s_width / 2) - (s_width / 3),
            sdiffs_plot[season],
            width=s_width,
            label=season[2:],
            facecolor=season_colours[season],
            edgecolor="k",
            lw=0.4,
        )
    ax.bar(
        years_plot + 0.5,
        annrf_dev.loc[years_plot],
        facecolor=(0, 0, 0, 0.1),
        width=1,
        lw=0.5,
        edgecolor=(0, 0, 0, 0.5),
        label="Annual",
    )
    ax.legend(loc="best", frameon=False, fontsize="small", ncol=4)
    ax.set_xticks(list(years_plot) + [years_plot[-1] + 1])
    plt.setp(ax.get_xticklabels(), rotation=90)
    ax.grid(True, color="gray", lw=0.2)
    y0, y1 = ax.get_ylim()
    dy = y1 - y0
    ax.set_ylim(y0, y1 + (dy / 12))
    ax.set_xlim(years_plot[0], years_plot[-1] + 1)
    ax.yaxis.set_major_locator(mticker.MaxNLocator(10))
    ax.set_ylabel(f"Difference from mean rainfall ({value_type})")
    ax.set_xlabel("Year (tick mark is located on Jan 1)")
    ax.set_title(title)
    return ax


def plot_seasonal_rainfall_totals_internal(
    station, report, daily_df=None, chart_start=2000, chart_end=None, fig_width=None
):
    if daily_df is None:
        ddf = report.get_rainfall_data(station, "daily")
    else:
        ddf = daily_df

    ddf["year"] = ddf.Date.dt.year
    monthly = reduce_daily_to_monthly(ddf, year_col="year")

    if chart_end is None:
        # chart_end = (station["dfn"].rf_period_finish - pd.Timedelta(days=365)).year
        chart_end = station["dfn"].rf_period_finish.year

    # title = f"{report.report_key} - seasonal totals - {station['station']} ({station['dfn'].station_name})"
    title = f"{station['id']} {station['name']}"

    monthly["season"] = monthly.month.map(months_to_seasons)
    monthly.loc[monthly.season == "summer", "season"] = "0-summer"
    monthly.loc[monthly.season == "autumn", "season"] = "1-autumn"
    monthly.loc[monthly.season == "winter", "season"] = "2-winter"
    monthly.loc[monthly.season == "spring", "season"] = "3-spring"
    monthly["year-season"] = monthly["year"].astype(str) + "-" + monthly["season"]
    monthly = monthly.sort_values(["year", "month"])
    seasons = (
        monthly.groupby(["year-season", "year", "season"]).Rain.sum().reset_index()
    )
    ann = seasons.groupby("year").Rain.sum()

    chart_sel = lambda x: x["year"] >= chart_start
    chart_sel_y = lambda x: (x["year"] - 1) >= chart_start
    ann = seasons.groupby("year").Rain.sum().reset_index()
    adf = ann[chart_sel]
    sdf = seasons[chart_sel]
    summer_sel = lambda x: x.season == "0-summer"
    autumn_sel = lambda x: x.season == "1-autumn"
    winter_sel = lambda x: x.season == "2-winter"
    spring_sel = lambda x: x.season == "3-spring"

    if fig_width is None:
        fig_width = 1 + (adf.year.max() - adf.year.min()) * 0.3
    fig = plt.figure(figsize=(fig_width, 5))
    ax2 = fig.add_subplot(111)
    ax = ax2.twinx()
    # ax2.bar(sdf[winter_sel].index, adf.Rain, width=3.7, color='k', alpha=0.2)
    ax2.bar(
        sdf[winter_sel].index - 1 / 6,
        adf.Rain[adf.year.isin(sdf[sdf.season == "2-winter"].year.unique())],
        width=4,
        color="darkblue",
        lw=0.7,
        facecolor="none",
        edgecolor="k",
        label="Annual (calendar)",
    )
    ax.bar(
        sdf.index,
        sdf.Rain.mask(sdf.season != "0-summer", np.nan),
        width=1,
        color="sandybrown",
        label="Summer",
    )
    ax.bar(
        sdf.index,
        sdf.Rain.mask(sdf.season != "1-autumn", np.nan),
        width=1,
        color="khaki",
        label="Autumn",
    )
    ax.bar(
        sdf.index,
        sdf.Rain.mask(sdf.season != "2-winter", np.nan),
        width=1,
        color="deepskyblue",
        label="Winter",
    )
    ax.bar(
        sdf.index,
        sdf.Rain.mask(sdf.season != "3-spring", np.nan),
        width=1,
        color="yellowgreen",
        label="Spring",
    )
    # ax.plot(sdf[winter_sel].index, sdf[summer_sel].Rain, lw=0.7, ls='-', marker='.', color='darkgoldenrod')
    # ax.plot(sdf[winter_sel].index, sdf[autumn_sel].Rain, lw=0.7, ls='-', marker='.', color='olivedrab')
    # ax.plot(sdf[winter_sel].index, sdf[winter_sel].Rain, lw=0.7, ls='-', marker='.', color='blue')
    # ax.plot(sdf[winter_sel].index, sdf[spring_sel].Rain, lw=0.7, ls='-', marker='.', color='purple')
    ax2.legend(loc="upper right", frameon=False, fontsize="small", ncol=1)
    ax.legend(loc="upper left", frameon=False, fontsize="small", ncol=2)
    ax2_y0, ax2_y1 = ax2.get_ylim()
    ax2.set_ylim(ax2_y0, ax2_y1 * 1.1)
    ax_y0, ax_y1 = ax.get_ylim()
    ax.set_ylim(ax_y0, ax_y1 * 1.5)
    years = sdf.loc[sdf.season == "2-winter", "year"]
    xts = ax2.set_xticks(years.index)
    xtls = ax2.set_xticklabels(years)
    xtls = plt.setp(ax2.get_xticklabels(), rotation=90)
    ax2.set_ylabel("Seasonal rainfall (mm)")
    ax.set_ylabel("Annual rainfall (mm)")
    ax.set_xlim(sdf.index[0], sdf.index[-1])
    ax.set_title(title)
    ax.set_xlabel("Year - tick mark is on Jan 1")
    return ax


def plot_rainfall_daily_intensity(daily_df, years, threshold=10):
    daily_df["month"] = daily_df.Date.dt.month
    months, seasons = get_seasonal_rainfall_data(daily_df)
    seasonal_lut = months.set_index(["year", "month"])["season"]
    for year, month in daily_df[["year", "month"]].drop_duplicates().to_numpy():
        daily_df.loc[
            (daily_df.year == year) & (daily_df.month == month), "season"
        ] = seasonal_lut.loc[(year, month)]
    fig = plt.figure(figsize=(7.2, 3.4))
    ax = fig.add_subplot(111)
    days = daily_df[daily_df.year.isin(years)]
    days_low = days[days.Rain <= threshold]
    days_high = days[days.Rain > threshold]
    months = (
        daily_df[daily_df.year.isin(years)]
        .groupby(["year", "month"], as_index=False)
        .Rain.sum()
    )
    months["left_day"] = months.apply(
        (lambda r: pd.Timestamp(year=int(r.year), month=int(r.month), day=1)),
        axis="columns",
    )
    # ax.bar(months.left_day, months.Rain, width=29, align='edge', facecolor)
    ax.bar(
        days_low.Date,
        days_low.Rain,
        width=1,
        facecolor="gray",
        edgecolor="gray",
        linewidth=0.5,
        label="Daily rainfall",
    )
    ax.bar(
        days_high.Date,
        days_high.Rain,
        width=1,
        facecolor="blue",
        edgecolor="blue",
        linewidth=0.5,
        label=f"Days above {threshold} mm",
    )
    ax.xaxis.set_minor_locator(mdates.MonthLocator(bymonth=np.arange(1, 13, 1)))
    ax.xaxis.set_major_locator(mdates.MonthLocator(bymonth=(5, 11), bymonthday=1))
    ax.xaxis.set_minor_formatter(mticker.NullFormatter())
    # ax.xaxis.set_minor_formatter(mdates.DateFormatter("%Y-%m-%d"))
    ax.xaxis.set_major_formatter(mdates.DateFormatter("%b %Y"))
    # major_xtls = plt.setp(ax.get_xticklabels())
    minor_xtls = plt.setp(ax.get_xticklabels(minor=True), fontsize="small", rotation=90)
    ax.set_xlim(
        pd.Timestamp(str(years[0])), pd.Timestamp(year=years[1], month=12, day=31)
    )
    ax.set_ylabel("Daily rainfall (mm)", fontsize="medium")
    leg = ax.legend(loc="upper left", fontsize="small", frameon=False, ncol=2)
    y0, y1 = ax.get_ylim()
    dy = y1 - y0
    ax.set_ylim(y0, y1 + dy / 3.5)
    for year in years:
        year_xloc = pd.Timestamp(year=year, month=7, day=1)
        count = len(
            days_high[
                (days_high.year == year)
                & (days_high.month >= 5)
                & (days_high.month <= 10)
            ]
        )
        if count == 0:
            label = "No days above 10 mm"
        elif count == 1:
            label = "1 day above 10 mm"
        else:
            label = f"{count:.0f} days above 10 mm"
        ax.text(
            year_xloc,
            y1,
            label + f"\nfrom May to Oct {year}",
            ha="center",
            va="bottom",
            color="darkblue",
            fontsize="small",
        )
    for spine in ["top", "right"]:
        ax.spines[spine].set_visible(False)
    return ax



import colorsys
from datetime import date, datetime, timedelta
import io
import re
import os
import logging
import textwrap
from pathlib import Path

import matplotlib.pyplot as plt
from matplotlib import lines as mlines
from matplotlib import gridspec
from matplotlib import dates as mdates
from matplotlib import ticker as mticker
from matplotlib import colors as mcolors
from matplotlib import patches as mpatches
import numpy as np
import pandas as pd
from adjustText import adjust_text
from scipy import stats
from PIL import Image, ImageChops
import pyproj

from .gwutils import *
from .utils import *
from .charts_utils import *

logger = get_logger()


def plot_wl_data_validation(
    wls,
    removals=None,
    well_ids=None,
    well_id_col="well_id",
    well_title_col="well_title",
    dt_col="obs_date",
    wl_col="rswl",
    label_col="reason",
    show_comments=True,
    adjust_comments=False,
    adjust_text_lim=5,
    path=".",
    ms=3,
    savefig_and_close=False,
    dpi=130,
    leg_fontsize="x-small",
    comment_fontsize="x-small",
    **kws,
):
    """Plot a graph of salinity data for validation purposes.

    Args:
        wls (pandas.DataFrame): water level data
        removals (pandas.DataFrame): removed water level data (should contain
            the column *label_col* as well)
        well_id_col (str): unique identifier for the well in both dataframes.
        well_ids (list): if you only want to plot one or two wells, specify
            them here. By default, it will iterate over all wells in the dataframes.
        dt_col (str): datetime column in both dataframes
        wl_col (str): water level data column in both dataframes
        ms (int): marker size default =3
        label_col (str): reason for removal of data (only needed in the
            *removals* dataframe)
        show_comments (bool): show comments from SA Geodata as text labels.
        adjust_comments (bool): use the adjustText python library to attempt
            and move the comment labels so that they do not overlap. It is quite
            slow, so be aware of that.
        adjust_text_lim (int): number of iterations for adjustText to run. 2 or 3
            will be fast but likely ineffective for all but the simplest plots.
            100 will be quite slow but a thorough test on almost all plots. 30
            is a good compromise.
        path (str): path to save figures into.
        savefig_and_close (bool): save fig to png and close
        dpi (int): resolution of saved PNGs00

    Returns:
        list: a list of matplotlib Axes.

    Any keyword arguments beginning with ``adjust_text`` will be sent to the
    :func:`adjustText.adjust_text` function, so for example to tell adjustText to
    use autoalign, pass *adjust_text_autoalign=True* to this function.

    .. todo:: Add example chart for plot_wl_data_validation

    """
    if well_ids is None:
        if removals is not None:
            if len(removals):
                well_ids = sorted(
                    list(
                        set(
                            list(wls[well_id_col].unique())
                            + list(removals[well_id_col].unique())
                        )
                    )
                )
        else:
            well_ids = list(set(list(wls[well_id_col].unique())))
    axes = []

    if savefig_and_close:
        fig = plt.figure(figsize=(8, 4))

    for i, well_id in enumerate(well_ids):
        logger.debug(f"Charting WL data for {well_id}")
        if savefig_and_close:
            ax = fig.add_subplot(111)
        else:
            fig = plt.figure(figsize=(8, 4))
            ax = fig.add_subplot(111)

        w_wls = wls[wls[well_id_col] == well_id]
        if not well_title_col in w_wls:
            well_title = ""
        else:
            well_title = w_wls[well_title_col].dropna()
            if len(well_title) > 1:
                well_title = well_title.iloc[0]
            else:
                well_title = well_id
        logger.debug(f"Also known as {well_title}")

        if removals is not None:
            w_removals = removals[removals[well_id_col] == well_id]

        ax.set_prop_cycle(color=removal_reason_colours)
        if removals is not None:
            for reason, rdf in w_removals.groupby("reason"):
                ax.plot(rdf[dt_col], rdf[wl_col], marker="X", lw=0, label=reason)
        if len(w_wls):
            ax.plot(
                w_wls[w_wls.database == "Aquarius"][dt_col],
                w_wls[w_wls.database == "Aquarius"][wl_col],
                lw=0,
                marker=".",
                ms=ms,
                color="gray",
                alpha=1,
                label="Aquarius",
            )
            ax.plot(
                w_wls[w_wls.database == "SA Geodata"][dt_col],
                w_wls[w_wls.database == "SA Geodata"][wl_col],
                lw=0,
                marker=".",
                ms=ms,
                color="black",
                alpha=1,
                label="SA Geodata",
            )
            (wl_line,) = ax.plot(
                w_wls[dt_col],
                w_wls[wl_col],
                lw=0.2,
                marker="",
                color="black",
                alpha=1,
                label="Filtered data",
            )
        texts = []
        if show_comments:
            w_wls_c = w_wls[~pd.isnull(w_wls.comments)]
            if len(w_wls_c):
                ax.plot(
                    w_wls_c[dt_col],
                    w_wls_c[wl_col],
                    marker="o",
                    lw=0,
                    mfc="none",
                    mec="purple",
                    mew=0.5,
                    label="SA Geodata comments",
                )
                for j, (idx, row) in enumerate(w_wls_c.iterrows()):
                    x = row[dt_col]
                    y = row[wl_col]
                    text = ax.text(
                        x,
                        y,
                        f'"{row.comments}"'.replace(".", ".\n"),
                        color="purple",
                        fontsize=comment_fontsize,
                    )
                    texts.append(text)
        ax.legend(fontsize=leg_fontsize, frameon=False, ncol=2)
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.set_title(well_title)
        ax.set_ylabel(wl_col)
        ax.xaxis.set_major_locator(mdates.AutoDateLocator())
        ax.xaxis.set_minor_locator(mdates.YearLocator())
        ax.grid(True, lw=0.5, color="#aaaaaa", alpha=0.5, ls="-")
        ax.grid(
            True, which="minor", axis="x", lw=0.5, color="#aaaaaa", alpha=0.5, ls=":"
        )

        y0, y1 = sorted(ax.get_ylim())
        if wl_col.lower() != "rswl":
            ax.set_ylim(y1, y0)
        else:
            ax.set_ylim(y0, y1)

        # Rearrange the labels if requested
        if adjust_comments:
            logger.debug(
                f"Using adjust_text to shift comment annotations for {well_id}"
            )
            adjust_text_kws = {}
            adjust_text_kws["arrowprops"] = kws.get(
                "adjust_text_arrowprops", dict(arrowstyle="->", color="purple", lw=0.4)
            )
            adjust_text_kws["autoalign"] = kws.get("adjust_text_autoalign", False)
            adjust_text_kws["lim"] = adjust_text_lim
            for key, argval in kws.items():
                if key.startswith("adjust_text"):
                    adjust_text_kws[key.replace("adjust_text_", "")] = argval
            adjust_text(
                texts,
                x=mdates.date2num(wl_line.get_xdata()),
                y=wl_line.get_ydata(),
                **adjust_text_kws,
            )

        if savefig_and_close:
            fig.savefig(
                str(Path(path) / f"plot_wl_data_validation_{well_id}.png"), dpi=dpi
            )
            fig.clf()
        else:
            axes.append(ax)

    return axes


def plot_wl_months_coloured(
    df, wl_col="rswl", dt_col="obs_date", mcmap=None, fig=None, ax=None, cax=None
):
    """Plot water level with months highlighted in colours.

    Args:
        df (:class:`pandas.DataFrame`): table of data to plot
        wl_col (str): column of *df* with water level measurements
        dt_col (str): column of *df* with datetime of observations
        mcmap (:class:`wrap_technote.MonthlyColormap`, optional):
            special object catering for the colour mapping to use
            for the months.
        fig (:class:`matplotlib.pyplot.Figure`)
        ax (:class:`matplotlib.pyplot.Axes`): axes to use for the
            main chart
        cax (:class:`matplotlib.pyplot.Axes`): axes to use for the
            colour mapping legend

    Returns:
        dict: dictionary with keys *ax* (:class:`matplotlib.pyplot.Axes`),
        *cbar* (:class:`matplotlib.colorbar.Colorbar`), and
        *mcmap* (:class:`wrap_technote.MonthlyColormap`)

    Example:

    .. code-block:: python

        >>> import wrap_technote, dew_gwdata
        >>> db = dew_gwdata.sageodata()
        >>> wls = db.water_levels(db.find_wells("PLL013"))
        >>> wrap_technote.plot_wl_months_coloured(wls)

    .. figure:: figures/plot_wl_months_coloured.png

    """
    if mcmap is None:
        mcmap = MonthlyColormap()
    if fig is None:
        fig = plt.figure(figsize=(7, 3))
    if ax is None:
        ax = fig.add_subplot(111)

    ax.plot(df[dt_col], df[wl_col], marker="None", lw=0.5, color="black")
    pc = ax.scatter(
        df[dt_col],
        df[wl_col],
        c=[mcmap.cmap(mcmap.norm(d.month)) for d in df[dt_col]],
        lw=0.1,
        edgecolor="black",
        marker="o",
        s=30,
        zorder=10,
        cmap=mcmap.cmap,
        norm=mcmap.norm,
        alpha=0.8,
    )
    cbar = plt.colorbar(mcmap.mappable, cax=cax)

    month_names = [
        "Jan",
        "Feb",
        "Mar",
        "Apr",
        "May",
        "Jun",
        "Jul",
        "Aug",
        "Sep",
        "Oct",
        "Nov",
        "Dec",
    ]
    c0, c1 = cbar.mappable.get_clim()
    cbar_ticks = []
    cbar_labels = []
    cbar_divisions = 12
    cbar_step = (c1 - c0) / cbar_divisions
    for i, value in enumerate(range(1, 13)):
        cbar_labels.append(month_names[i])
        cbar_ticks.append(((cbar_step * i) + c0) + (cbar_step / 2))

    cbar.set_ticks(cbar_ticks)
    cbar.set_ticklabels(cbar_labels)

    ax.set_ylabel(wl_col)

    y0, y1 = sorted(ax.get_ylim())
    if wl_col.lower() != "rswl":
        ax.set_ylim(y1, y0)
    else:
        ax.set_ylim(y0, y1)

    return {"ax": ax, "cbar": cbar, "mcmap": mcmap}


def plot_wl_seasonality(
    df,
    anndf=None,
    seasons=None,
    dots=False,
    annmarkers="large & open",
    dt_col="obs_date",
    wl_col="rswl",
    span_length=5,
    style="default",
):
    """Make a chart for analysing seasonal highs and lows for a well.

    Args:
        df (:class:`pandas.DataFrame`): water level data
        anndf (:class:`pandas.DataFrame`, optional): annual seasonal
            water level max and mins. Should
            also have a season column "season", and each season for each year,
            "year+season", each containing strings.
            If it is not supplied, no seasonal breakdown is charted.
        seasons (:class:`wrap_technote.Seasons`, optional): definition of seasons
        dots (bool): show each measurement as a dot
        annmarkers (str): markers for the annual mins and maxs. Either "large & open"
            or "small & filled".
        dt_col (str): datetime column in ``df`` (and ``anndf`` if supplied)
        wl_col (str): water level column in ``df`` (and ``anndf`` if supplied)
        span_length (int): number of years in each pane of the figure
        style (str): matplotlib style

    Returns:
        dict: dictionary with keys "fig" and "axes" (the latter is a list)

    Example:

    .. code-block:: python

        >>> import wrap_technote, dew_gwdata
        >>> db = dew_gwdata.sageodata()
        >>> wls = db.water_levels(db.find_wells("MOR100"))
        >>> wrap_technote.plot_wl_seasonality(wls, dots=True)

    .. figure:: figures/plot_wl_seasonality_1.png

    To define seasonal summer/recovery periods and also show the seasonal
    max and mins:

    .. code-block:: python

        >>> seasons = (
        ...     wrap_technote.Seasons()
        ...     .append("recovery", "max", end=wrap_technote.doy("Jan 15"))
        ...     .append("pumped", "min", end=wrap_technote.doy("Jul 1"))
        ...     .append("recovery", "max")
        ... )
        >>> annualdata = wrap_technote.analyse_wl_by_seasons(wls, seasons)
        >>> wrap_technote.plot_wl_seasonality(wls, annualdata, seasons=seasons)

    .. figure:: figures/plot_wl_seasonality_2.png

    """

    colours = [
        np.asarray(plt.cm.hsv(i)) * 0.8 for i in np.linspace(0, 1, span_length + 1)
    ]

    year_min = df[dt_col].dt.year.min()
    year_max = df[dt_col].dt.year.max()

    yearspan_starts = list(get_yearspans_between(year_min, year_max + 1, span_length))
    dayofyears = list(month_dayofyears.values())
    month_labels = list(month_dayofyears.keys())

    # Obtain a grid of chart axes.
    ncol = 3
    nrow = int(np.ceil((len(yearspan_starts) / ncol)))
    logger.debug(f"style on plot_wl_seasonality: {style}")
    with plt.style.context(style):
        fig, axes = plt.subplots(
            nrow,
            ncol,
            figsize=(10, 3.0 * nrow),
            sharey="row",
            sharex=True,
            gridspec_kw=dict(wspace=0.05, hspace=0),
        )

        # Get the Axes object as a 1-dimensional list.
        axes = axes.ravel()

        ylims = {}
        for k, yearspan_start in enumerate(yearspan_starts):
            col_ix = k % ncol

            # Use a new axis for each
            ax = axes[k]

            yearspan_end = yearspan_start + span_length

            plotted = False

            # Chart each year as a separate line.
            for i, year in enumerate(range(yearspan_start, yearspan_end)):
                # We want the line to extend off the chart into the previous year's data,
                # and to the following year's data. Any gap larger than 3 years will appear
                # as a break in the line.
                df3yr = df[abs(df[dt_col].dt.year - year) <= 1]
                doy_offset = (df3yr[dt_col] - datetime(year, 1, 1)).dt.days

                # Ensure that January 1st (i.e. offset of 0) has value of 1
                doy_offset += 1

                # If there is data from the year in question, chart the line.
                if dots:
                    marker = "o"
                    ms = 1
                else:
                    marker = "None"
                    ms = None
                if colours:
                    c = colours[i]
                else:
                    c = None
                if len(df3yr[df3yr[dt_col].dt.year == year]):
                    plotted = True
                    (l2,) = ax.plot(
                        doy_offset,
                        df3yr[wl_col],
                        lw=0.5,
                        marker=marker,
                        ms=ms,
                        alpha=1,
                        label=year,
                        color=c,
                    )

                if not anndf is None and seasons is not None:
                    # Find any maxmins occurring in this calendar year and plot them
                    dfym = anndf[anndf[dt_col].dt.year == year]
                    for season, dfyms in dfym.groupby("season"):
                        s_doy_offset = (
                            dfyms[dt_col] - datetime(year, 1, 1)
                        ).dt.days + 1
                        marker = seasons.period_kws[season].get("marker", "s")
                        if annmarkers == "large & open":
                            kws = dict(ms=9, mec=l2.get_color(), mew=1, mfc="none")
                        elif annmarkers == "small & filled":
                            kws = dict(
                                ms=4, mec=l2.get_color(), mew=1, mfc=l2.get_color()
                            )
                        ax.plot(
                            s_doy_offset,
                            dfyms[wl_col],
                            marker=marker,
                            ls="none",
                            label="",
                            **kws,
                        )

            labels_plotted = []
            if not seasons is None:
                # Show the seasons as striped sections.
                for period in seasons.periods:
                    season = period["season"]
                    color = seasons.period_kws[season].get("color", "k")
                    if season in labels_plotted:
                        label = None
                    else:
                        label = season
                    ax.axvspan(
                        period["start"],
                        period["end"],
                        color=color,
                        alpha=0.05,
                        label=label,
                    )
                    labels_plotted.append(season)

            ax.legend(fontsize="x-small", ncol=3, frameon=True, framealpha=0.4)
            ax.set_xlim(0, 366)
            ax.set_xticks(dayofyears)
            ax.set_xticklabels(month_labels)
            if k < ncol:
                ax.tick_params(
                    axis="x", bottom=True, top=True, labelbottom=False, labeltop=True
                )
            plt.setp(
                ax.get_xticklabels(), fontsize="x-small"
            )  # rotation=20, ha="right")
            ax.grid(True, lw=0.1, alpha=0.2)
            if k % ncol == 0:
                ax.set_ylabel(wl_col, fontsize="small")
            plt.setp(ax.get_yticklabels(), fontsize="small")

        for ka in range(len(axes)):
            if ka > k:
                axes[ka].set_visible(False)
            else:
                if wl_col.lower() != "rswl":
                    w_bottom = df[wl_col].max()
                    w_top = df[wl_col].min()
                else:
                    w_bottom = df[wl_col].min()
                    w_top = df[wl_col].max()
                axes[ka].set_ylim(w_bottom, w_top)

        # fig.tight_layout()
        return {"fig": fig, "axes": axes}


def plot_wl_seasonal_timeseries(
    df, anndf, scatter=False, seasons=None, mcmap=None, ax=None, fig=None, cax=None
):
    """Plot the seasonal max and min water levels for a time series.

    Args:
        df (:class:`pandas.DataFrame`): water level data set
        anndf (:class:`pandas.DataFrame`): data set of annual max and mins
            from *df* (use :func:`wrap_technote.analyse_wl_by_seasons` to
            obtain this table)
        scatter (bool): show annual max/mins as dots
        seasons (:class:`wrap_technote.Seasons`): definition of seasonal
            variations, useful for controlling the plotting style of the
            max/min lines. Not needed normally.
        mcmap (:class:`wrap_technote.MonthlyColormap`, optional): show
            coloured dots for the month of each observation.
        ax (Axes, optional): main plot
        fig (Figure, optional): main figure
        cax (matplotlib Axes, optional): axes to draw the color mapping legend
            in (only needed if scatter is True and mcmap is supplied)

    Returns:
        :class:`matplotlib.Axes`: axes object

    Example:

    .. code-block:: python

        >>> import wrap_technote, dew_gwdata
        >>> db = dew_gwdata.sageodata()
        >>> wls = db.water_levels(db.find_wells("PLL013"))
        >>> wrap_technote.plot_wl_months_coloured(wls, dots=True, span_length=6)
        >>> # Define seasonal periods
        >>> seasons = (
        ...     wrap_technote.Seasons()
        ...     .append("summer", "min", end=doy("2018-03-15"), marker="v", color="red")
        ...     .append("recovery", "max", end=doy("2018-11-15"), marker="^", color="blue")
        ...     .append("summer", "min")
        ... )
        >>> # Calculate annual max and mins
        >>> annualdata = wrap_technote.analyse_wl_by_seasons(wls, seasons)
        >>> wrap_technote.plot_wl_seasonal_timeseries(wls, annualdata)

    .. figure:: figures/plot_wl_seasonal_timeseries_1.png

    To change the way the chart looks - e.g. show recovery as a black line
    and to not show summer minimums.

    .. code-block:: python

        >>> seasons.period_kws["recovery"]["color"] = "black"
        >>> seasons.period_kws["summer"]["color"] = "none"
        >>> wrap_technote.plot_wl_seasonal_timeseries(
        ...     wls,
        ...     annualdata,
        ...     seasons=seasons,
        ...     scatter=True,
        ...     mcmap=wrap_technote.MonthlyColormap(),
        ... )

    .. figure:: figures/plot_wl_seasonal_timeseries_2.png

    """
    if fig is None:
        fig = plt.figure(figsize=(7, 3))
    if ax is None:
        ax = fig.add_subplot(111)

    if scatter:
        if mcmap:
            c = [mcmap.norm(d.month) for d in df.obs_date]
            cmap = mcmap.cmap
        else:
            c = "gray"
            cmap = None
        cset = ax.scatter(
            df.obs_date,
            df.rswl,
            c=c,
            lw=0.1,
            edgecolor="black",
            cmap=cmap,
            label="",
            s=12,
        )

        if mcmap:
            cbar = plt.colorbar(mcmap.mappable, cax=cax)

            month_names = [
                "Jan",
                "Feb",
                "Mar",
                "Apr",
                "May",
                "Jun",
                "Jul",
                "Aug",
                "Sep",
                "Oct",
                "Nov",
                "Dec",
            ]
            c0, c1 = cbar.get_clim()
            cbar_ticks = []
            cbar_labels = []
            cbar_divisions = 12
            cbar_step = (c1 - c0) / cbar_divisions
            for i, value in enumerate(range(1, 13)):
                cbar_labels.append(month_names[i])
                cbar_ticks.append(((cbar_step * i) + c0) + (cbar_step / 2))

            cbar.set_ticks(cbar_ticks)
            cbar.set_ticklabels(cbar_labels)

    ax.plot(df.obs_date, df.rswl, marker="None", color="gray", lw=0.4, label="All data")

    for season, seasondf in anndf.groupby("season"):
        if not seasons is None:
            color = seasons.period_kws[season]["color"]
        else:
            color = None
        ax.plot(
            seasondf.obs_date,
            seasondf.rswl,
            label=season,
            lw=0.8,
            color=color,
            marker=None,
        )

    ax.legend(fontsize="small")
    ax.set_ylabel("RSWL (m AHD)")
    return ax


def plot_wl_bom_classes(
    df_classes,
    df_all=None,
    dt_col="obs_date",
    wl_col="rswl",
    bom_class_col="rswl_bom_class",
    ax=None,
    fig=None,
    cax=None,
    bcmap=None,
    classes_scatter=None,
    all_scatter=None,
    break_hydrograph_at=None,
    tight_layout=True,
):
    """Create hydrograph with annual ranked recovered levels highlighted
    using the BoM classification.

    Args:
        df_classes (:class:`pandas.DataFrame`): table with the BoM
            classification of the ranked recovered water level
            (see a filtered version of :func:`wrap_technote.analyse_wl_by_seasons`
            for how to obtain this table)
        df_all (:class:`pandas.DataFrame`): table with all data
            for the background hydrograph
        dt_col (str): column in both *df_classes* and *df_all* with the
            measurement datetime
        wl_col (str): column in both *df_classes* and *df_all* with the
            annual/recovered water level
        bom_class_col (str): column in *df_classes* with the ranked BoM
            classification as a string e.g. "Very much below average"
        ax (mpl.Axes)
        fig (mpl.Figure)
        cax (either False, None, or mpl.Axes): if it is False, then a
            colorbar will not be drawn.
        bcmap (:class:`wrap_technote.BoMClassesColormap`)
        break_hydrograph_at (tuple): list of datetimes to break the df_all table into to
            plot.

    Returns:
        dict: dictionary with keys *ax* (:class:`matplotlib.pyplot.Axes`),
        *cbar* (:class:`matplotlib.colorbar.Colorbar`), and
        *bcmap* (:class:`wrap_technote.MonthlyColormap`)

    Example:

    .. code-block:: python

        >>> import wrap_technote as tn
        >>> import dew_gwdata
        >>> db = dew_gwdata.sageodata()
        >>> wls = db.water_levels(db.find_wells("BRS020"))
        >>> wls = tn.filter_wl_observations(wls)
        >>> seasons = (
        ...     tn.Seasons()
        ...     .append("stressed", "min", end=tn.doy("May 15"))
        ...     .append("recovery", "max", end=tn.doy("Nov 15"))
        ...     .append("stressed", "min")
        ... )
        >>> wls_ann = tn.analyse_wl_by_seasons(wls, seasons)
        >>> # Keep only the recovered annual water level.
        >>> wls_rec = wls_ann.loc[wls_ann.season == "recovery"]
        >>> wls_rec["rswl_percentile"] = wls_rec.rswl.rank(pct=True) * 100
        >>> wls_rec["rswl_bom_class"] = wls_rec.rswl_percentile.transform(tn.percentile_to_bom_class)
        >>> tn.plot_wl_bom_classes(wls_rec, wls)

    .. figure:: figures/plot_wl_bom_classes.png

    """
    if ax is None:
        if fig is None:
            fig = plt.figure(figsize=(7, 3))
        ax = fig.add_subplot(111)
    if bcmap is None:
        bcmap = BoMClassesColormap()
    if classes_scatter is None:
        classes_scatter = {}
    if all_scatter is None:
        all_scatter = {}
    all_scatter = dict(dict(lw=0.3, color="black"), **all_scatter)
    classes_scatter = dict(
        dict(edgecolor="k", linewidth=0.5, s=50, marker="o"), **classes_scatter
    )
    if break_hydrograph_at is None:
        break_hydrograph_at = []
    break_hydrograph_at.append(df_all[dt_col].max() + timedelta(days=1))

    if df_all is not None:
        start_date = df_all[dt_col].min() - timedelta(days=1)
        for end_date in sorted(break_hydrograph_at):
            didx = (df_all[dt_col] > start_date) & (df_all[dt_col] < end_date)
            ax.plot(df_all.loc[didx, dt_col], df_all.loc[didx, wl_col], **all_scatter)
            start_date = end_date
    cset = ax.scatter(
        df_classes[dt_col],
        df_classes[wl_col],
        c=[bom_classes.index(x) for x in df_classes[bom_class_col]],
        cmap=bcmap.cmap,
        **classes_scatter,
    )
    cbar = None
    if cax is not False:
        cbar = plt.colorbar(cset, cax=cax)
        bcmap.fix_ticklabels(cbar)

    ax.set_xlabel("Date")
    ax.set_ylabel("RSWL (m AHD)")
    if tight_layout:
        ax.figure.tight_layout()

    y0, y1 = sorted(ax.get_ylim())
    if wl_col.lower() != "rswl":
        ax.set_ylim(y1, y0)
    else:
        ax.set_ylim(y0, y1)

    return {"ax": ax, "cax": cax, "cbar": cbar}


def plot_wls_with_logger(
    manual_df,
    logger_df=None,
    ax=None,
    dt_col="obs_date",
    wl_col="dtw",
    zoom_to="all",
    zoom_buffer=None,
):
    """Plot a graph of water levels including Aquarius logger data
    if available.

    Args:
        manual_df (:class:`pandas.DataFrame`): manual WL observations
        logger_df (:class:`pandas.DataFrame`): continuous logger WL
            observations
        dt_col (str): columns in both *manual_df* and *logger_df*
            that contain datetimes
        wl_col (str): columns in both *manual_df* and *logger_df*
            that contain water levels (measured as depth to water)
        zoom_to (str): "all", "logger", "since-logger"
        zoom_buffer (timedelta): buffer on zoom, 1 month by default
            You can set it to False for no buffer

    Returns:
        :class:`matplotlib.pyplot.Axes`: axes object

    Example:

    .. code-block:: python

        >>> import wrap_technote as tn, dew_gwdata
        >>> db = dew_gwdata.sageodata()
        >>> sag_df = db.water_levels(db.find_wells("MUW034"))
        >>> hyd_df = db.fetch_hydstra_dtw_data(wells)
        >>> tn.plot_wls_with_logger(sag_df, hyd_df)

    .. figure:: figures/plot_wls_with_logger.png

    The *zoom_to* keyword argument will help with visualizing wells that
    have a long history:

    .. code-block:: python

        >>> sag_df = db.water_levels(db.find_wells("YAT37"))
        >>> hyd_df = db.fetch_hydstra_dtw_data(wells)
        >>> tn.plot_wls_with_logger(sag_df, hyd_df, zoom_to="since-logger")

    .. figure:: figures/plot_wls_with_logger_2.png

    """
    if ax is None:
        ax = plt.gca()
    if not logger_df is None:
        for qc, dfq in logger_df.groupby("grade"):
            m = "."
            (f,) = ax.plot(dfq[dt_col], dfq[wl_col], lw=0.8)
            ax.scatter(
                [],
                [],
                marker=m,
                facecolor=f.get_color(),
                s=100,
                label=f"Aquarius QC\n{qc}",
            )
        if len(logger_df):
            lims = (
                max(logger_df[wl_col].max(), manual_df[wl_col].max()),
                min(logger_df[wl_col].min(), manual_df[wl_col].min()),
            )
        else:
            lims = manual_df[wl_col].max(), manual_df[wl_col].min()
        if wl_col.lower() == "rswl":
            lims = lims[::-1]
        ax.set_ylim(*lims)
        ax.set_ylabel(wl_col)
    ax.plot(
        manual_df[dt_col],
        manual_df[wl_col],
        marker="o",
        mec="k",
        mfc="none",
        label="Manually-dipped",
        lw=0.5,
        mew=0.3,
        color="k",
        ls=":",
        ms=2,
    )
    if zoom_buffer is None:
        zoom_buffer = timedelta(days=30)
    elif zoom_buffer is False:
        zoom_buffer = timedelta(seconds=1)
    if len(logger_df):
        if zoom_to == "logger":
            ax.set_xlim(
                logger_df[dt_col].min() - zoom_buffer,
                logger_df[dt_col].max() + zoom_buffer,
            )
        elif zoom_to == "since-logger":
            ax.set_xlim(logger_df[dt_col].min() - zoom_buffer, None)
    ax.legend(loc="best", fontsize="small")
    return ax


def plot_wl_trend(
    wls,
    trend_wls,
    trend_lines=None,
    trend_line_start=None,
    trend_line_end=None,
    all_wls=None,
    well_id_col="well_id",
    well_title_col="well_id",
    dt_col="obs_date",
    wl_col="rswl",
    hist_min_date_col="rswl_min_date",
    ax=None,
    ms=5,
    year_span=12,
    override_year_span=False,
):
    """Plot water level trend.

    Args:
        wls (pd.DataFrame):
        trend_wls (pd.DataFrame):
        trend_lines (?):
        trend_line_start (?):
        trend_line_end (?):
        all_wls (pd.DataFrame, optional):
        well_id_col (str): column with the well identifier
        well_title_col (str): column with the well title
        dt_col (str): column(s) containing dates
        wl_col (str): column containing water levels - either 'rswl', 'swl', or 'dtw'
        hist_min_date_col (str): TODO
        ax (matplotlib.pyplot.Axes)
        ms (int)
        year_span (int)
        override_year_span (?)

    Returns:
        :class:`matplotlib.pyplot.Axes`: axes object

    .. todo:: Document plot_wl_trend function, add example code and figure

    """
    well_id = wls[well_id_col].unique()[0]
    if ax is None:
        fig = plt.figure(figsize=(8, 3))
        ax = fig.add_subplot(111)
    if all_wls is not None:
        ax.plot(
            all_wls[dt_col],
            all_wls[wl_col],
            lw=0.5,
            marker="o",
            ms=ms - 2,
            mew=0.2,
            mec="grey",
            mfc="grey",
            color="grey",
            alpha=0.2,
            label="All data",
        )
    lw = 0.5
    ax.plot(
        wls[dt_col],
        wls[wl_col],
        color="#222222",
        lw=lw,
        marker="o",
        mfc="orange",
        mec="gray",
        ms=ms - 1,
        mew=0.3,
        alpha=0.5,
        label="Annual recovered WL",
    )
    ax.plot(
        trend_wls[dt_col],
        trend_wls[wl_col],
        marker="o",
        mfc="red",
        mec="k",
        ms=ms,
        mew=0.3,
        lw=0,
        alpha=0.7,
        label="Used for trend calculation",
    )
    hist_idx = wls[dt_col] == trend_wls[hist_min_date_col].iloc[0]
    ax.plot(
        wls[hist_idx][dt_col],
        wls[hist_idx][wl_col],
        marker="o",
        mec="purple",
        mfc="none",
        mew=0.5,
        ms=ms + 5,
        ls="none",
        label="Historical minimum",
    )

    if trend_lines is not None:
        trend_line = trend_lines.loc[well_id]
        trend_line_ts = [trend_line_start, trend_line_end]
        trend_line_wls = calculate_trendline_at_dates(trend_line, trend_line_ts)
        ax.plot(trend_line_ts, trend_line_wls, lw=1, color="r", label="Linear trend")
        end_ts = trend_line_ts[1]
        fg, bg = status_to_colours(trend_line.status_change, param="WL")
        trend_label = f": {trend_line.status_change} at {trend_line.slope_yr:.2f} m/y"
    else:
        end_ts = pd.Timestamp(trend_wls[dt_col].max())
        fg = "black"
        bg = "white"

    start_ts = end_ts - timedelta(days=365.25 * year_span)
    if override_year_span:
        hist_min_ts = trend_wls[hist_min_date_col].iloc[0] - timedelta(days=120)
        if hist_min_ts < start_ts:
            start_ts = hist_min_ts
    ax.set_xlim(start_ts, end_ts + timedelta(days=120))
    leg = ax.legend(loc="best", fontsize="small", frameon=False)
    if year_span <= 15:
        ax.xaxis.set_major_locator(mdates.YearLocator(1))
    ax.xaxis.set_major_formatter(mdates.DateFormatter("%Y"))

    ax.set_ylabel(wl_col)
    title_text = ax.set_title(
        f"{wls[well_title_col].iloc[0]}{trend_label}",
        color=fg,
        backgroundcolor=bg,
        fontsize="medium",
    )
    if all_wls is not None:
        lim_wls_whole = all_wls
    else:
        lm_wls_whole = wls
    lim_wls = lim_wls_whole[
        (lim_wls_whole[dt_col] >= start_ts) & (lim_wls_whole[dt_col] <= end_ts)
    ][wl_col]
    yspan = lim_wls.max() - lim_wls.min()
    y0 = lim_wls.min() - yspan * 0.1
    y1 = lim_wls.max() + yspan * 0.1
    if wl_col == "rswl":
        ax.set_ylim(y0, y1)
    else:
        ax.set_ylim(y1, y0)

    ax._title_text = title_text
    return ax


def plot_wl_rankings_internal(curr_ranks, title=""):
    """Plot bar chart of WL rankings for internal use.

    Args:
        curr_ranks (DataFrame): e.g. from
            `resource.read_data("recovery_wl_data", "current ranked WLs")
        title (str): chart title e.g. resource_key

    Returns:
        :class:`matplotlib.pyplot.Figure`: figure object

    The figure shows the number and percent of wells for each
    BoM class in both bar chart and stacked bar chart form,
    with the median well's class indicated on the figure.

    .. todo:: Add example code and figure for plot_wl_rankings_internal

    """

    bcmap = BoMClassesColormap()
    bom_keys = bcmap.class_names
    class_counts = bom_classes_dict()
    class_counts.update(curr_ranks.rswl_bom_class.value_counts().to_dict())

    fig = plt.figure(figsize=(6, 3))
    gs = gridspec.GridSpec(1, 2, width_ratios=(5, 2))
    ax = fig.add_subplot(gs[0])
    ax2 = fig.add_subplot(gs[1])

    bc_range = [i for i in range(len(bom_keys))]
    bc_counts = [class_counts[x] for x in bom_keys]
    ax.barh(bc_range, bc_counts, color=bcmap.colours_rgba(1), edgecolor="k", lw=0.8)
    ax.set_yticks(range(len(bcmap.labels)))
    _ = ax.set_yticklabels(bcmap.labels, fontsize="small")
    label_ypos = np.asarray(bc_counts) + (max(bc_counts) / 30)
    percentage_counts = [y / sum(bc_counts) * 100 for y in bc_counts]
    percentage_labels = [
        f"{pct:.0f}" for pct in round_to_100_percent(percentage_counts, 0)
    ]
    bottom = 0
    for i in range(len(bc_range)):
        if bc_counts[i] > 0:
            ax.text(
                label_ypos[i],
                bc_range[i],
                f"{bc_counts[i]} wells ({percentage_labels[i]}%)",
                va="center",
                ha="left",
                color="#222222",
                fontsize="small",
            )
        ax2.bar(
            [1],
            [percentage_counts[i]],
            bottom=bottom,
            facecolor=bcmap.class_to_rgba(bom_keys[i]),
            edgecolor="k",
            lw=0.5,
        )
        bottom += percentage_counts[i]
    ax.set_xlabel("Number of wells", fontsize="medium")
    y0, y1 = ax.get_xlim()
    ax.xaxis.set_major_locator(
        mticker.MaxNLocator(nbins="auto", steps=[1, 2, 5, 10], integer=True)
    )
    ax.set_xlim(0, y1 + np.ceil((max(bc_counts) / 30)))
    for sp in ["top", "right"]:
        ax.spines[sp].set_visible(False)
    for sp in ["top", "bottom", "left"]:
        ax2.spines[sp].set_visible(False)
    ax2.tick_params(
        axis="y",
        left=False,
        labelleft=False,
        right=True,
        labelright=True,
    )
    ax2.set_xticks([])
    ax2.set_xlim(0, 1.5)
    ax2.set_ylim(-0.5, 100)
    ax2.set_ylabel(
        f"Percentage of wells ({sum(bc_counts)} wells in total)", fontsize="small"
    )
    ax2.yaxis.set_label_position("right")
    median_ranking = get_median_ranking(curr_ranks["rswl_bom_class"])
    ax2.annotate(
        f"Median well: {median_ranking}",
        (1.4, 50),
        (0.4, 50),
        arrowprops=dict(arrowstyle="->", color="k"),
        color="k",
        rotation=90,
        ha="right",
        va="center",
        fontsize="small",
    )

    ax.set_title(title, fontsize="medium")
    fig.tight_layout()
    return fig


def plot_wl_historical_rankings(ranks, qc, title="", figsize=None):
    """Plot historical recovered WLs

    Args:
        ranks (pd.DataFrame): all ranked WLs
        qc (pd.DataFame): data quality table
        title (str)
        figsize (tuple): width, height in inches

    Returns:
        :class:`matplotlib.pyplot.Figure`: matplotlib Figure

    """
    figsize_override = {}

    # curr_wells = ranks[ranks.obs_date.dt.year == 2019].well_id.unique()
    # ranks = ranks[ranks.well_id.isin(curr_wells)]

    year_min = ranks.season_year.str[:4].astype(int).min()
    year_max = ranks.season_year.str[:4].astype(int).max()
    years = np.arange(year_min, year_max + 1, step=1)
    ncols = len(years)
    year_labels = []
    for i in range(ncols):
        if i < len(years):
            year_labels.append(f"{years[i]:.0f}")
        else:
            year_labels.append("")

    unsorted_well_ids = ranks.well_id.unique()
    sort_records = []
    for w in unsorted_well_ids:
        qc_condition = qc[qc.well_id == w].all_conditions.iloc[0]
        years_data = len(
            ranks[ranks.well_id == w].season_year.str[:4].astype(int).unique()
        )
        sort_records.append(
            {"well_id": w, "qc_condition": qc_condition, "years_data": years_data}
        )
    sort_records = (
        pd.DataFrame(sort_records)
        .sort_values(["qc_condition", "years_data"], ascending=False)
        .reset_index(drop=True)
    )

    well_ids = sort_records.well_id.values
    qc_false = sort_records[sort_records.qc_condition == False].index
    if len(qc_false):
        false_condition = qc_false[0]
    else:
        false_condition = np.nan
    n_wells = len(well_ids)

    bcmap = BoMClassesColormap()
    ranks_arr = np.empty((n_wells, ncols), dtype=int)
    for i, well_id in enumerate(well_ids):
        for j in range(ncols):
            if j < len(years):
                year = years[j]
            ranking = None
            ranks_subset = ranks[
                (ranks.well_id == well_id)
                & (ranks.season_year.str[:4].astype(int) == year)
            ]
            if len(ranks_subset):
                ranking = ranks_subset.iloc[0].rswl_bom_class
                short_ranking = bcmap.class_names.index(ranking)
            else:
                short_ranking = -1
            ranks_arr[i, j] = short_ranking

    logger.debug(f"Number of wells: {len(well_ids)}")
    logger.debug(f"Number of years: {len(years)} ({year_min} to {year_max})")

    if figsize is None:
        figsize = (len(years) / 6.5, len(well_ids) / 5.6)
        logger.debug(f"Figure size in inches: {figsize}")
        if figsize[0] < 3.5:
            figsize = (3.5, figsize[1])
            logger.debug(f"Overriding figure width. figsize now {figsize}")
        if figsize[1] < 4:
            figsize = (figsize[0], 4)
            logger.debug(f"Overriding figure height. figsize now {figsize}")

    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    ax.imshow(ranks_arr, cmap=bcmap.cmap_nodata)

    yts = ax.set_yticks(range(n_wells))
    ytls = ax.set_yticklabels(well_ids, fontsize="xx-small")
    ax.tick_params(
        "y", which="major", left=True, right=True, labelleft=True, labelright=True
    )

    xts = ax.set_xticks(range(len(year_labels)))
    xtls = ax.set_xticklabels(year_labels, fontsize="x-small", rotation=90, ha="center")
    ax.tick_params(
        "x", which="major", bottom=True, top=True, labelbottom=True, labeltop=True
    )

    ax.axhline(false_condition - 0.5, color="green", lw=1)
    for year in years:
        if year % 10 == 0:
            year_label = f"{year:.0f}"
            if year_label in year_labels:
                dashes = (1, 0)
                if year_label == "2000":
                    lw = 1
                    ls = "-"
                elif year_label == "2010":
                    lw = 1
                    ls = "--"
                    dashes = (10, 3)
                else:
                    lw = 0.5
                    ls = ":"
                ax.axvline(
                    year_labels.index(year_label) - 0.5,
                    color="black",
                    lw=lw,
                    ls=ls,
                    dashes=dashes,
                )

    ax.set_xticks(np.arange(-1, len(year_labels)) + 0.5, minor=True)
    ax.set_yticks(np.arange(-1, n_wells) + 0.5, minor=True)
    ax.grid(True, which="minor", axis="both", lw=0.2, color="grey")
    ax.tick_params("x", which="minor", bottom=False)
    ax.tick_params("y", which="minor", left=False)
    ax.set_title(title)
    ax.grid(False)
    fig.tight_layout()
    return fig


def plot_wl_trends_internal(wltrends, title=""):
    return plot_trends_internal(
        wltrends,
        title=title,
        class_names=("Declining", "Stable", "Rising"),
        param_yaxis_label="m/y",
        param_col="slope_yr",
    )


def plot_wl_ranking_classes(curr_ranks, ax=None):
    """Plot_wl_ranking_classes

    Args:
        curr_ranks (?)

    Returns:
        :class:`matplotlib.pyplot.Axes`: axes

    .. todo::

        Document the plot_wl_ranking_classes function. This is
        the one that goes into the Tech Note.

    """
    bcmap = BoMClassesColormap()
    bom_keys = bcmap.class_names
    class_counts = bom_classes_dict()
    class_counts.update(curr_ranks.rswl_bom_class.value_counts().to_dict())
    if ax is None:
        fig = plt.figure(figsize=(3.6, 2.8))
        ax = fig.add_subplot(111)

    bc_range = [i for i in range(len(bom_keys))]
    bc_counts = [class_counts[x] for x in bom_keys]
    ax.barh(bc_range, bc_counts, color=bcmap.colours_rgba(1), edgecolor="k")
    ax.set_yticks(range(len(bcmap.labels)))
    _ = ax.set_yticklabels(bcmap.labels, fontsize="small")
    label_ypos = np.asarray(bc_counts) + (max(bc_counts) / 30)
    percentage_counts = [y / sum(bc_counts) * 100 for y in bc_counts]
    percentage_labels = [
        f"{pct:.0f}" for pct in round_to_100_percent(percentage_counts, 0)
    ]
    for i in range(len(bc_range)):
        if bc_counts[i] > 0:
            ax.text(
                label_ypos[i],
                bc_range[i],
                percentage_labels[i] + "%",
                va="center",
                ha="left",
                color="#555555",
            )
    ax.set_xlabel("Number of wells", fontsize="medium")
    y0, y1 = ax.get_xlim()
    ax.xaxis.set_major_locator(
        mticker.MaxNLocator(nbins="auto", steps=[1, 2, 5, 10], integer=True)
    )
    ax.set_xlim(0, y1 + np.ceil((max(bc_counts) / 30)))
    for sp in ["top", "right"]:
        ax.spines[sp].set_visible(False)

    return ax


def plot_wl_ranking_map(
    curr_ranks,
    map_elements,
    map_annotations,
    ax=None,
    leg_frame=True,
    map_crs=8059,
    leg_loc="lower right",
    markersize=45,
):
    """plot_wl_ranking_map

    Args:
        curr_ranks
        map_elements
        map_annotations
        ax
        leg_frame (bool)
        map_crs (int)
        leg_loc (str)
        markersize (int)

    Returns:
        :class:`matplotlib.pyplot.Axes`: axes

    .. todo::

        Document the plot_wl_ranking_map function. This is
        the one that goes into the Tech Note.

    """
    if ax is None:
        fig = plt.figure(figsize=(3.6, 2.8))
        ax = fig.add_subplot(111)

    for spine in ax.spines.values():
        spine.set_visible(False)

    bcmap = BoMClassesColormap()
    bom_keys = bcmap.class_names

    legend = []
    if len(map_elements):
        if "legend_zorder" in map_elements[0]:
            map_elements = sorted(map_elements, key=lambda x: x["legend_zorder"])
    for me in map_elements:
        me["shapely_object"].plot(ax=ax, **me["plot_kwargs"])
        if me.get("label", None):
            if me["legend_item_type"] == "line":
                legend_patch = ax.plot([], [], **me["legend_kwargs"])[0]
            elif me["legend_item_type"] == "rect":
                legend_patch = mpatches.Rectangle((0, 0), 1, 1, **me["legend_kwargs"])
            legend.append((me["label"], legend_patch))

    tr = pyproj.Transformer.from_crs(7844, map_crs, always_xy=True)
    for bom_class in bom_keys[::-1]:
        dfg = curr_ranks[curr_ranks["rswl_bom_class"] == bom_class]
        if len(dfg):
            easting, northing = tr.transform(dfg.longitude.values, dfg.latitude.values)
            ax.scatter(
                easting,
                northing,
                marker="o",
                color=bcmap.class_to_rgba(bom_class),
                edgecolor="#555555",
                linewidth=0.3,
                s=markersize,
                zorder=5,
            )

    for ann in map_annotations:
        logger.debug(f"annotating map with {ann}")
        if ann["well_id"] in curr_ranks.well_id.unique():
            row = curr_ranks[curr_ranks.well_id == ann["well_id"]].iloc[0]
            label = ann.get("label", ann["well_id"])
            arrowprops = ann.get("arrowprops", None)
            if arrowprops is not None:
                arrowprops = dict(
                    {
                        "arrowstyle": "-|>",
                        "color": "#444444",
                        "mutation_scale": 8,
                        "shrinkA": 0,
                        "shrinkB": 0,
                    },
                    **arrowprops,
                )
            easting, northing = tr.transform(row.longitude, row.latitude)
            ax.annotate(
                label,
                (easting, northing),
                (
                    easting + ann.get("x_km", 0) * 1000,
                    northing + ann.get("y_km", 0) * 1000,
                ),
                color="#444444",
                fontsize="x-small",  # map annotation
                ha=ann.get("ha", "left"),
                va=ann.get("va", "bottom"),
                arrowprops=arrowprops,
                zorder=6,
            )
        else:
            logger.debug(f"{ann['well_id']} not found in {curr_ranks.well_id.unique()}")
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_xticklabels([])
    ax.set_yticklabels([])

    if len(legend):
        labels, patches = zip(*legend)
        if leg_frame:
            frameon = True
            framealpha = 1
            edgecolor = "white"
        else:
            frameon = False
            framealpha = None
            edgecolor = None
        leg = ax.legend(
            patches,
            labels,
            loc=leg_loc,
            fontsize="x-small",
            frameon=frameon,
            framealpha=framealpha,
            edgecolor=edgecolor,
        )
        for t in leg.get_texts():
            t.set_color("#444444")
    return ax


def plot_wl_trend_triclass_bars(wl_trend_df, ax=None):
    """todo

    Args:
        wl_trend_df (pd.DataFrame)
        ax

    Returns:
        :class:`matplotlib.pyplot.Axes`: Axes object

    """
    gsr_colours = status_change_colours
    gsr_colours_rgba = {
        k: np.asarray([x / 255 for x in rgb] + [1]) for k, rgb in gsr_colours.items()
    }
    status_key_x = [0, 1, 2]
    status_keys = ["Rising", "Stable", "Declining"]
    status_values = [
        len(wl_trend_df[wl_trend_df.status_change == v]) for v in status_keys
    ]
    status_colours = [gsr_colours_rgba[s] for s in status_keys]

    if ax is None:
        fig = plt.figure(figsize=(3.6, 2.0))
        ax = fig.add_subplot(111)

    ax.barh(
        status_key_x,
        status_values,
        height=0.6,
        color=status_colours,
        edgecolor="#555555",
    )
    ax.set_yticks(status_key_x)
    ax.set_yticklabels(status_keys)
    ax.set_ylim(status_key_x[-1] + 1, status_key_x[0] - 1)

    label_ypos = np.asarray(status_values) + (max(status_values) / 30)
    percentage_counts = [y / sum(status_values) * 100 for y in status_values]
    percentage_labels = [
        f"{pct:.0f}" for pct in round_to_100_percent(percentage_counts, 0)
    ]
    for i, x in enumerate(status_key_x):
        if percentage_counts[i] > 0:
            ax.text(
                label_ypos[i],
                x,
                percentage_labels[i] + "%",
                va="center",
                ha="left",
                color="#555555",
            )
    ax.set_xlabel("Number of wells", fontsize="medium")

    max_counts = max(status_values)
    ax.set_xlim(0, max_counts + (max_counts / 2.0))
    ax.xaxis.set_major_locator(mticker.MaxNLocator(integer=True, steps=[1, 2, 5, 10]))

    for sp in ("top", "right"):
        ax.spines[sp].set_visible(False)
    return ax


def plot_wl_trend_triclass_map(
    wl_trend_df,
    map_elements,
    map_annotations,
    ax=None,
    leg_frame=True,
    map_crs=8059,
    leg_loc="lower right",
    markersize=60,
):
    wl_trend_map_markers = {"Rising": "^", "Stable": "s", "Declining": "v"}

    if ax is None:
        fig = plt.figure(figsize=(3.6, 2.8))
        ax = fig.add_subplot(111)

    for spine in ax.spines.values():
        spine.set_visible(False)

    gsr_colours = status_change_colours
    gsr_colours_rgba = {
        k: np.asarray([x / 255 for x in rgb] + [1]) for k, rgb in gsr_colours.items()
    }

    legend = []
    if len(map_elements):
        if "legend_zorder" in map_elements[0]:
            map_elements = sorted(map_elements, key=lambda x: x["legend_zorder"])
    for me in map_elements:
        me["shapely_object"].plot(ax=ax, **me["plot_kwargs"])
        if me.get("label", None):
            if me["legend_item_type"] == "line":
                legend_patch = ax.plot([], [], **me["legend_kwargs"])[0]
            elif me["legend_item_type"] == "rect":
                legend_patch = mpatches.Rectangle((0, 0), 1, 1, **me["legend_kwargs"])
            legend.append((me["label"], legend_patch))

    tr = pyproj.Transformer.from_crs(7844, map_crs, always_xy=True)
    for trend_label in ["Rising", "Stable", "Declining"]:
        dfg = wl_trend_df[wl_trend_df.status_change == trend_label]
        if len(dfg):
            easting, northing = tr.transform(dfg.longitude.values, dfg.latitude.values)
            ax.scatter(
                easting,
                northing,
                marker=wl_trend_map_markers[trend_label],
                color=gsr_colours_rgba[trend_label],
                edgecolor="#555555",
                linewidth=0.3,
                s=markersize,
                zorder=5,
            )

    logger.debug(f"wl_trend_df: \n{wl_trend_df}")

    for ann in map_annotations:
        logger.debug(f"Charting annotation: {ann}")
        if ann["well_id"] in wl_trend_df.well_id.unique():
            row = wl_trend_df[wl_trend_df.well_id == ann["well_id"]].iloc[0]
            label = ann.get("label", ann["well_id"])
            arrowprops = ann.get("arrowprops", None)
            if arrowprops is not None:
                arrowprops = dict(
                    {
                        "arrowstyle": "-|>",
                        "color": "#444444",
                        "mutation_scale": 8,
                        "shrinkA": 0,
                        "shrinkB": 0,
                    },
                    **arrowprops,
                )
            easting, northing = tr.transform(row.longitude, row.latitude)
            ax.annotate(
                label,
                (easting, northing),
                (easting + ann["x_km"] * 1000, northing + ann["y_km"] * 1000),
                color="#444444",
                fontsize="x-small",  # map annotation
                ha=ann["ha"],
                va=ann["va"],
                arrowprops=arrowprops,
                zorder=6,
            )
        else:
            logger.debug(
                f" > did not find {ann['well_id']} in {wl_trend_df.well_id.unique()}"
            )
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_xticklabels([])
    ax.set_yticklabels([])

    if len(legend):
        labels, patches = zip(*legend)
        if leg_frame:
            frameon = True
            framealpha = 1
            edgecolor = "white"
        else:
            frameon = False
            framealpha = None
            edgecolor = None
        leg = ax.legend(
            patches,
            labels,
            loc=leg_loc,
            fontsize="x-small",
            frameon=frameon,
            framealpha=framealpha,
            edgecolor=edgecolor,
        )
        for t in leg.get_texts():
            t.set_color("#444444")
    return ax



from setuptools import setup

setup(
    name="wrap_technote",
    packages=["wrap_technote"],
    use_scm_version={"version_scheme": "post-release"},
    setup_requires=["setuptools_scm"],
    description="Code for the annual WRAP technical notes",
    long_description=open("README.md", mode="r").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/dew-waterscience/wrap_technote",
    author="DEW Water Science (Kent Inverarity)",
    author_email="kent.inverarity@sa.gov.au",
    license="All rights reserved @ DEW",
    classifiers=(
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering",
    ),
    keywords="science groundwater",
    install_requires=(
        "loguru",
        "matplotlib",
        "shapely",
        "pyshp",
        "jinja2",
        "click",
        "requests",
        "python-sa-gwdata>=0.5.4",
        "adjustText",
        "xlrd",
        "openpyxl",
        "pillow",  # conda
        "pyproj",
        "geopandas>=0.5.0",  # conda
        "scipy",
        "attrdict",
        "ausweather",
        "sageodata_db>=0.19",
        "dew_gwdata>=0.74",
    ),
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "wraptn = wrap_technote.scripts.cli:wraptn",
            "wrap_run_waterlevels = wrap_technote.scripts.waterlevels:waterlevels",
            "wrap_run_salinities = wrap_technote.scripts.salinities:salinities",
            "wrap_run_rainfall = wrap_technote.scripts.rainfall:rainfall",
            "wrap_run_summaries = wrap_technote.scripts.summaries:summaries",
        ]
    },
)



import collections
from pathlib import Path
import logging
import io
import sqlite3

# import datatest
import numpy as np
import pandas as pd
import requests

import ausweather

# import pytest
import wrap_technote as tn


from .utils import *

logger = get_logger()


class RainfallStationData:
    """Rainfall station data.

    You should initialise this by using any of these class methods:

    - :meth:`wrap_technote.RainfallStationData.from_bom_via_silo`
    - :meth:`wrap_technote.RainfallStationData.from_aquarius`
    - :meth:`wrap_technote.RainfallStationData.from_wrap_report`
    - :meth:`wrap_technote.RainfallStationData.from_data`

    e.g.

    .. code-block::

        >>> rf = tn.RainfallStationData.from_bom_via_silo('18017')

    You can then access data via the ``rf.daily``, ``rf.calendar``, or
    ``rf.financial`` attributes.

    Args:
        station_id (str): station ID
        exclude_incomplete_years (bool): only show complete years

    """

    def __init__(self, station_id, exclude_incomplete_years=False):
        self.station_id = station_id
        self.exclude_incomplete_years = False

    @property
    def exclude_incomplete_years(self):
        return self.__exclude_incomplete_years

    @exclude_incomplete_years.setter
    def exclude_incomplete_years(self, value):
        if value:
            self.__exclude_incomplete_years = True
        else:
            self.__exclude_incomplete_years = False

    @classmethod
    def from_bom_via_silo(cls, station_id, data_start=None):
        """Create from BoM data (via SILO).

        Args:
            station_id (str): BoM Station ID
            data_start (pd.Timestamp): date to download data from, default 1/1/1950

        Returns:
            :class:`wrap_technote.RainfallStationData`

        Note that this will download the data afresh from the SILO website.

        """
        if data_start is None:
            data_start = pd.Timestamp("1950-01-01")
        self = cls(station_id)
        self.df = download_bom_rainfall(station_id, data_start)
        return self

    @classmethod
    def from_aquarius(cls, station_id, data_start=None):
        """Create from Aquarius Timeseries data (for South Australia only)

        Args:
            station_id (str): AQTS LocationIdentifier.
            data_start (pd.Timestamp): date to download data from, default 1/1/1950

        Returns:
            :class:`wrap_technote.RainfallStationData`

        Note that this will download the data afresh from AQTS.

        """
        if data_start is None:
            data_start = pd.Timestamp("1950-01-01")
        self = cls(station_id)
        self.df = download_aquarius_rainfall(station_id, data_start)

        return self

    @classmethod
    def from_wrap_report(cls, station_id, report):
        """Create from WRAP Report dataset.

        Args:
            report (:class:`wrap_technote.Report`)
            station_id (str):

        Returns:
            :class:`wrap_technote.RainfallStationData`

        This will not download data, but instead (attempt to)
        retrieve it from the already-downloaded WRAP report
        SQLite database. The download would have occurred when
        somebody ran the ``wraptn rainfall ...`` command-line
        script.

        """
        self = cls(station_id)
        dfn = report.rainfall_dfn(self.station_id)
        data = report.get_rainfall_data(id=station_id, sheet=None)
        self.df = data["daily"]
        return self

    @property
    def daily(self):
        """Daily rainfall data.

        Returns:
            :class:`pandas.DataFrame`: dataframe with columns:

            - date (pd.Timestamp)
            - rainfall (float)
            - interpolated_code (int)
            - quality (int)
            - year (int)
            - dayofyear (int)
            - finyear (str)
            - station_id (str)

        """
        df = self.df.assign(station_id=self.station_id)
        return df

    @property
    def calendar(self):
        df = self.groupby("year").assign(station_id=self.station_id)
        df.insert(1, "start_date", [pd.Timestamp(f"{y}-01-01") for y in df.year])
        if self.exclude_incomplete_years:
            missing = tn.find_missing_days(
                self.daily, dt_col="date", year_type="calendar", value_col="rainfall"
            )
            complete = missing[missing == 0]
            df = df[df.year.isin(complete.index.values)]
        return df

    @property
    def financial(self):
        df = self.groupby("finyear").assign(station_id=self.station_id)
        df.insert(1, "start_date", [pd.Timestamp(f"{y[:4]}-07-01") for y in df.finyear])
        if self.exclude_incomplete_years:
            missing = tn.find_missing_days(
                self.daily, dt_col="date", year_type="financial", value_col="rainfall"
            )
            complete = missing[missing == 0]
            df = df[df.finyear.isin(complete.index.values)]
        return df

    def groupby(self, grouping_column):
        """Group daily rainfall by either calendar or financial year.

        Args:
            grouping_column (str): either 'year' or 'finyear'

        Returns:
            :class:`pandas.DataFrame`: dataframe with these columns:

            - year or finyear (str)
            - rainfall (float): rainfall in mm
            - rainfall_count (int): number of days with data
            - interpolated_count (int): number of days with interpolated_code != 0
            - quality_count (int): number of days with non-null quality code.

        """
        assert grouping_column in self.df.columns
        return self.df.groupby(grouping_column, as_index=False).agg(
            rainfall=("rainfall", "sum"),
            rainfall_count=("rainfall", "count"),
            interpolated_count=(
                "interpolated_code",
                lambda x: len([xi for xi in x if xi != 0]),
            ),
            quality_count=(
                "quality",
                lambda x: len([xi for xi in x if not pd.isnull(xi)]),
            ),
        )

    def save_to_excel(self, report=None, fn=None, overwrite=True):
        """Save data to Excel.

        Args:
            report (:class:`wrap_technote.Report`): used to derive the filename path
            fn (str): you can pass this directly if you wish
                overwrite (bool)

        """
        d = convert_daily_to_old_style_rainfall_sheets(self.daily)

        if fn is None:
            dfn = report.rainfall_dfn(self.station_id)
            stub = str(self.station_id) + "_" + dfn.station_name
            excel_filename = (
                report.rainfall_path / f"aquarius_data_{report.report_key}_{stub}.xlsx"
            )
        else:
            excel_filename = Path(fn)
        if overwrite:
            with pd.ExcelWriter(excel_filename) as writer:
                d["df"].to_excel(writer, "daily", index=False)
                d["annual"].reset_index().to_excel(
                    writer, "annual (calendar)", index=False
                )
                d["srn"].reset_index().rename(columns={"year": "Date"}).to_excel(
                    writer, "annual provenance", index=False
                )
                d["wateruse_year"].reset_index().to_excel(
                    writer, "annual (water-use year)", index=False
                )

    def save_to_database(self, report):
        """Save data to a SQLite database in the rainfall folder of a WRAP report.

        Args:
            report (:class:`wrap_technote.Report`): used to derive the filename path

        """
        fn = report.rainfall_path / f"rainfall_data_{report.report_key}.db"
        # with sqlite3.connect(fn) as db:
        from sqlalchemy import create_engine

        db = create_engine("sqlite:///" + str(fn))  # C:\\path\\to\\foo.db")
        if 1:
            for table in ("daily", "calendar", "financial"):
                logger.info(f"Saving '{table}'")
                try:
                    dbdf = pd.read_sql(f"select * from {table}", db)
                    dbdf = dbdf[~(dbdf.station_id == str(self.station_id))]
                    new_dbdf = pd.concat([dbdf, getattr(self, table)])
                    logger.info(f"Joined existing stations to {self.station_id}")
                except:
                    new_dbdf = getattr(self, table)
                    logger.info(f"Nothing in database (error) - using df direct.")
                finally:
                    if "date" in new_dbdf.columns:
                        new_dbdf["date"] = pd.to_datetime(new_dbdf["date"]).dt.strftime(
                            "%Y-%m-%d"
                        )
                    if "date_start" in new_dbdf.columns:
                        new_dbdf["date_start"] = pd.to_datetime(
                            new_dbdf["date_start"]
                        ).dt.strftime("%Y-%m-%d")
                    if "start_date" in new_dbdf.columns:
                        new_dbdf["start_date"] = pd.to_datetime(
                            new_dbdf["start_date"]
                        ).dt.strftime("%Y-%m-%d")
                    for i, column in enumerate(new_dbdf.columns):
                        series = new_dbdf[column]
                        example = series.iloc[0]
                        logger.info(
                            f"Column {i} {column} dtype = {series.dtype} {type(example)}"
                        )
                    new_dbdf.to_sql(table, db, index=False, if_exists="replace")


def download_bom_rainfall(station_id, data_start):
    """Download BoM rainfall data from SILO.

    Args:
        station_id (int or str): BoM station ID
        data_start (pd.Timestamp): date to download data from

    Returns:
        dict: Has four keys: 'df', 'annual', 'srn', and 'wateruse_year'

    This function uses :func:`ausweather.fetch_bom_station_from_silo`
    in the background.

    """
    station_id = int(f"{float(station_id):.0f}")
    logger.debug(f"Downloading {station_id} from {data_start}")

    data = ausweather.fetch_bom_station_from_silo(
        station_id, "groundwater@sa.gov.au", data_start
    )

    df = data["df"]
    df["date"] = pd.to_datetime(df["Date"])
    # df["Grade"] = pd.NA
    df["Grade"] = 1
    df = df.rename(
        columns={
            "Rain": "rainfall",
            "Srn": "interpolated_code",
            "Grade": "quality",
        }
    )
    df["year"] = df["date"].dt.year
    df["dayofyear"] = df["date"].dt.dayofyear
    df["finyear"] = [tn.date_to_wateruseyear(d) for d in df["date"]]

    cols = [
        "date",
        "rainfall",
        "interpolated_code",
        "quality",
        "year",
        "dayofyear",
        "finyear",
    ]
    return df[cols]


def download_aquarius_rainfall(station_id, data_start=None):
    """Download rainfall data from DEW's Aquarius Web Portal website.

    Args:
        station_id (int or str): Aquarius location ID
        data_start (pd.Timestamp): date to download data from (optional)

    Returns:
        dict:  with four keys: 'df', 'annual', 'srn', and 'wateruse_year'

    """
    if data_start is None:
        data_start = pd.Timestamp("1800-01-01")

    logger.debug(f"Downloading {station_id} from {data_start} from Aquarius")

    url = f"https://water.data.sa.gov.au/Export/BulkExport?DateRange=EntirePeriodOfRecord&TimeZone=9.5&Calendar=CALENDARYEAR&Interval=Daily&Step=1&ExportFormat=csv&TimeAligned=True&RoundData=True&IncludeGradeCodes=True&IncludeApprovalLevels=False&IncludeQualifiers=False&IncludeInterpolationTypes=False&Datasets[0].DatasetName=Rainfall.Best%20Available--Continuous%40{station_id}&Datasets[0].Calculation=Aggregate&Datasets[0].UnitId=89"
    resp = requests.get(url, verify=False)

    buffer = io.StringIO()
    buffer.write(resp.text)
    buffer.seek(0)
    i = 0
    for line in buffer:
        if i == 2:
            station_name = line.strip("\n").strip(",").strip()
        break

    df = pd.read_csv(
        buffer, skiprows=5, names=["Date", "end_timestamp", "Rain", "Grade"]
    )

    df = df.rename(columns={"Date": "date", "Rain": "rainfall", "Grade": "quality"})
    df["interpolated_code"] = 0
    df["date"] = pd.to_datetime(df["date"])
    df["year"] = df["date"].dt.year
    df["dayofyear"] = df["date"].dt.dayofyear
    df["finyear"] = [tn.date_to_wateruseyear(d) for d in df["date"]]

    cols = [
        "date",
        "rainfall",
        "interpolated_code",
        "quality",
        "year",
        "dayofyear",
        "finyear",
    ]
    return df[cols]


def convert_daily_to_old_style_rainfall_sheets(df):
    """Convert daily rainfall data to the old style of sheets.

    Args:
        df (pd.DataFrame): columns should be 'date', 'rainfall',
            'interpolated_code', 'quality', 'year', 'dayofyear',
            and 'finyear'.

    Returns:
        dict:

    """
    df = df.rename(columns={"date": "Date", "rainfall": "Rain", "quality": "Grade"})
    df = df.dropna(subset=["Date"], how="any")
    df["Date"] = pd.to_datetime(pd.to_datetime(df.Date).dt.date)
    df["year"] = df.Date.dt.year
    df["wu_year"] = df.Date.apply(tn.date_to_wateruseyear)
    df = df.dropna(subset=["Rain"], how="any")
    df = df[df.Grade >= 1]
    df["Srn"] = 0
    df = df[["Date", "year", "wu_year", "Rain", "Srn", "Grade"]]

    annual = df.groupby(df.Date.dt.year).Rain.sum()

    all_dates = pd.date_range(start=df.Date.min(), end=df.Date.max(), freq="1d")
    df_all = df.set_index("Date").reindex(all_dates).reset_index()
    df_all["Date"] = df_all["index"]
    df_all["year"] = df_all.Date.dt.year
    df_all["wu_year"] = df_all.Date.apply(tn.date_to_wateruseyear)
    df_all["Srn"] = df_all.Srn.fillna(25)
    df_all["Grade"] = df_all.Grade.fillna(-15)
    df_all["Grade"] = df_all.Grade.astype(int)
    df_all["Srn"] = df_all.Srn.astype(int)
    df_all.loc[df_all.Srn.isin([25, 35, 75]), "srn"] = "Interpolated"
    df_all.loc[df_all.Srn.isin([13, 15]), "srn"] = "Deaccumulated"
    srn = df_all.groupby(["year", "srn"]).Date.count().unstack(level=1)
    srn = srn.reindex(sorted(df_all.year.unique()), fill_value=0)
    for column in ["Interpolated", "Deaccumulated"]:
        if not column in srn.columns:
            srn[column] = 0

    wateruse_year = df.groupby(df.wu_year, as_index=False).Rain.sum()

    return {
        "df": df,
        "annual": annual,
        "srn": srn,
        "wateruse_year": wateruse_year,
    }


def reduce_daily_to_monthly(
    daily_df, dt_col="Date", year_col="wu_year", value_col="Rain"
):
    """Reduce daily rainfall totals into monthly totals per year.

    Args:
        daily_df (pandas DataFrame): daily rainfall data
        dt_col (str): column of *daily_df* with the date as a datetime
        year_col (str): column of *daily_df* with a value to group as
            years. It could be the year itself i.e.
            `daily_df[dt_col].dt.year` or it could be the financial year
        value_col (str): column with the rainfall total

    Returns:
        pd.DataFrame: a dataframe with columns *year_col*, "month", and *value_col*.

    """
    grouper = daily_df.groupby([daily_df[year_col], daily_df[dt_col].dt.month])
    sums = grouper[value_col].sum()
    return sums.reset_index().rename(columns={dt_col: "month"})


def get_seasonal_rainfall_data(daily, dt_col="Date", year_col="year", value_col="Rain"):
    """Convert daily rainfall data to monthly and seasonal.

    Args:
        daily (pd DataFrame): daily rainfall totals

    Returns:
        tuple: tuple of length two: monthly, seasons (both pd.DataFrames)

    """
    daily["year"] = daily[dt_col].dt.year
    monthly = reduce_daily_to_monthly(daily, year_col=year_col)

    monthly["season"] = monthly.month.map(months_to_seasons)
    monthly.loc[monthly.season == "summer", "season"] = "0-summer"
    monthly.loc[monthly.season == "autumn", "season"] = "1-autumn"
    monthly.loc[monthly.season == "winter", "season"] = "2-winter"
    monthly.loc[monthly.season == "spring", "season"] = "3-spring"
    monthly["year-season"] = monthly["year"].astype(str) + "-" + monthly["season"]
    monthly = monthly.sort_values(["year", "month"])
    seasons = (
        monthly.groupby(["year-season", "year", "season"])[value_col]
        .sum()
        .reset_index()
    )
    return monthly, seasons
